#ifndef __CHANNEL_H__
#define __CHANNEL_H__
/*
   
	ORIGINAL CHAMELEON CODE COMMENTS - HAVE REMOTE RELEVANCE TO THIS HEADER
	
   Sending and receiving packets

   Packets are sent and received on connections.  In order to simultaneously
   provide a good fit with conventional message-passing systems and with 
   other more direct systems (e.g., sockets), I've defined a set of
   connection macros that are here translated into Chameleon message-passing
   calls.  These are somewhat complicated by the need to provide access to
   non-blocking operations

   These are not yet fully integrated into the code.  

   This file is designed for use with vendor message-passing systems through
   the Chameleon definitions.  Other systems should REPLACE this file.
   See mpid/ch_tcp and mpid/ch_shmem for examples.  Note also that once
   NewDevice creates a device, the local mpid.h is not modified, so that
   changes to packets.h can be accomplised by #defines

   In addition, we provide a simple way to log the "channel" operations
   If MPID_TRACE_FILE is set, we write information on the operation (both
   start and end) to the given file.  In order to simplify the code, we
   use the macro MPID_TRACE_CODE(name,channel).  Other implementations
   of channel.h are encourage to provide the trace calls; note that as macros,
   they can be completely removed at compile time for more 
   performance-critical systems.

 */

/* Temporary place for debug version of malloc/free stuff */

#ifndef __DLL_CODE__
	#ifndef WINMPICHAPI	
	#define WINMPICHAPI __declspec(dllimport)
	#endif
#else
	#ifndef WINMPICHAPI	
	#define WINMPICHAPI __declspec(dllexport)
	#endif
#endif


#define __MALLOC(ptr, type, space) \
	if (space <= 0) {\
		fprintf (stderr, "Attempt to allocate non-positive number of bytes.\n");\
		fprintf (stderr, "Size: [%d] source file %s, line %d.\n", space, __FILE__, __LINE__);\
		fprintf (stderr, "Aborting program...\n");\
		_ASSERTE(FALSE);\
	};\
	ptr = (type*) malloc ((space) * sizeof(type));\
	if (0 == ptr) {\
		fprintf (stderr, "Could not allocate a block of memory.\n");\
		fprintf (stderr, "Source file %s, line %d.\n", __FILE__, __LINE__);\
		fprintf (stderr, "Aborting program...\n");\
		_ASSERTE(FALSE);\
	};

	#define __FREE(ptr) \
	if (0 == ptr) {\
		fprintf (stderr, "Attempt to free NULL pointer.\n");\
		fprintf (stderr, "Source file %s, line %d.\n", __FILE__, __LINE__);\
		_ASSERTE(FALSE);\
	};\
	free (ptr);\
	ptr = (void*)0;
	
	#ifndef __MEMCPY
		#define __MEMCPY(dest, source, size) \
		if ((0 == dest)||(0 == source)) {\
			fprintf (stderr, "Attempt to copy memory block to or from NULL address.\n");\
			fprintf (stderr, "Source file %s, line %d.\n", __FILE__, __LINE__);\
			_ASSERTE(FALSE);\
		};\
		if (size <= 0 ) {\
			fprintf (stderr, "Attempt to copy non-positive number of bytes in memcpy.\n");\
			fprintf (stderr, "Size: [%d]; source file %s, line %d.\n", size, __FILE__, __LINE__);\
			_ASSERTE(FALSE);\
		};\
		memcpy (dest, source, size);
	#endif


/* Check partner macro */

#define CHECK_PARTNER(rank) \
	if (rank >= MPID_WorldSize) {\
		fprintf (stderr, "Invalid rank.\n");\
		_ASSERTE(FALSE);}


/* Global Constants */
extern int MPID_WorldSize;
extern int MPID_MyWorldRank;
extern MPID_WINNT_SHMEM_BUFFER_T** WINNT_ShmemBufs;

extern WINMPICHAPI HANDLE hEndProcessEvent;
extern HANDLE hTerminatorEvent;

/*
extern HANDLE hPacketReadyToSend;
extern HANDLE hPacketAvailable;
*/
/* Initialization function prototypes */

WINMPICHAPI void *MPID_WINNT_init (int*, char***);
WINMPICHAPI void *MPID_WINNT_WorldSize(int*);
WINMPICHAPI void *MPID_WINNT_MyRank(int*);
WINMPICHAPI void MPID_WINNT_finalize(void);
WINMPICHAPI void MPID_WINNT_Barrier(void);
WINMPICHAPI void MPID_WINNT_abort(void);

DWORD WINAPI SHMEMReceiveThread(LPVOID);
DWORD WINAPI SHMEMSendThread(LPVOID);
DWORD WINAPI GLOBALTerminatorThread(LPVOID lpV);


/* Function prototypes for the very basic chanel operations */

int  MPID_ControlMsgAvail (void);
void MPID_RecvAnyControl (MPID_PKT_T*, int, int*);
void MPID_SendControl (MPID_PKT_T*, int, int);
void MPID_SendControlBlock (MPID_PKT_T*, int, int);
void MPID_RecvFromChannel (void*, int, int);
void MPID_SendChannel (void*, int, int);

/* If someone wants to use blocking SendControlBlock ... */
#ifdef MPID_USE_SEND_BLOCK
	void MPID_SendControlBlock (MPID_PKT_T*, int, int);
#endif

/* Because a common operation is to send a control block, and decide whether
   to use SendControl or SendControlBlock based on whether the send is 
   non-blocking, we include a definition for it here: 
 */
#ifdef MPID_USE_SEND_BLOCK
	#define MPID_SENDCONTROL(mpid_send_handle,pkt,len,dest) \
		if (mpid_send_handle->is_non_blocking) {\
			MPID_SendControl( pkt, len, dest );}\
		else {\
			MPID_SendControlBlock( pkt, len, dest );}
#else
	#define MPID_SENDCONTROL(mpid_send_handle,pkt,len,dest) \
		MPID_SendControl( (MPID_PKT_T*)pkt, len, dest )
#endif


/* 
   Non-blocking versions (NOT required, but if PI_NO_NRECV and PI_NO_NSEND
   are NOT defined, they must be provided)
 */

#ifndef PI_NO_NRECV 

	#define MPID_IRecvFromChannel( buf, size, channel, id ) \

	#define MPID_WRecvFromChannel( buf, size, channel, id ) \

	#define MPID_RecvStatus( id ) \

	#define MPID_CancelRecvChannel( id ) \

#endif

/* Note that these use the tag based on the SOURCE, not the channel
   See MPID_SendChannel */

#ifndef PI_NO_NSEND

	#define MPID_ISendChannel( buf, size, channel, id ) \

	#define MPID_WSendChannel( buf, size, channel, id ) \

	/* Test the channel operation */
	#define MPID_TSendChannel( id ) \

	#define MPID_CancelSendChannel( id ) \

#endif

/* If nonblocking sends are defined, the MPID_SendData command uses them;
   otherwise, the blocking version is used.
   These rely on dmpi_send_handle and mpid_send_handle 
 */
#ifndef PI_NO_NSEND
	#define MPID_SendData( buf, size, channel, mpid_send_handle ) \
		if (mpid_send_handle->is_non_blocking) {\
			MPID_ISendChannel( address, len, dest, mpid_send_handle->sid );\
			dmpi_send_handle->completer=MPID_CMPL_SEND_NB;\
		}\
		else \
			{\
			mpid_send_handle->sid = 0;\
			MPID_SendChannel( address, len, dest );\
			DMPI_mark_send_completed( dmpi_send_handle );\
		}
#else
	#define MPID_SendData( buf, size, channel, mpid_send_handle ) \
		mpid_send_handle->sid = 0;\
		MPID_SendChannel( address, len, dest );\
		DMPI_mark_send_completed( dmpi_send_handle );
#endif

/*
   We also need an abstraction for out-of-band operations.  These could
   use transient channels or some other operation.  This is essentially for
   performing remote memory operations without local intervention; the need
   to determine completion of the operation requires some sort of handle.
   Here are the ones that we've chosen. Rather than call them transient 
   channels, we define "transfers", which are split operations.  Both 
   receivers and senders may create a transfer.

   Note that the message-passing version of this uses the 'ready-receiver'
   version of the operations.

   There is a problem with the receive transfer definition.  The simplest
   form would be MPID_TestRecvTransfer( dmpi_recv_handle->dev_rhandle.rid )
   However, it might be that that test only indicates whether the transfer
   if READY for completion, not that it HAS been completed.  It may require
   an additional step to actually complete the transfer, using more 
   information than just the rid.  For this reason, there is an additional
   MPID_CompleteRecvTransfer( ... ) called only when a test succeeds.
   Some implementations may leave this empty
   
 */

#define MPID_CreateSendTransfer( buf, size, partner, id ) {*(id) = 0;}
#define MPID_CreateRecvTransfer( buf, size, partner, id ) \
       {*(id) = CurTag++;TagsInUse++;}

#ifndef PI_NO_NRECV

	#define MPID_StartRecvTransfer( buf, size, partner, id, rid ) \
		{MPID_TRACE_CODE("BIRRRecv",id);\
		 PInrecvrr( MPID_PT2PT2_TAG(id), buf, size, MSG_OTHER, rid );\
		 MPID_TRACE_CODE("EIRRRecv",id);}
	#define MPID_EndRecvTransfer( buf, size, partner, id, rid ) \
		{MPID_TRACE_CODE("BIWRRecv",id);\
		 PIwrecvrr( MPID_PT2PT2_TAG(id), buf, size, MSG_OTHER, rid );\
		 MPID_TRACE_CODE("EIWRRecv",id);\
		 if (--TagsInUse == 0) CurTag = 1024; else if (id == CurTag-1) CurTag--;}
	#define MPID_TestRecvTransfer( rid ) \
		PInstatus( rid )
	#define MPID_CompleteRecvTransfer( buf, size, partner, id, rid )

#else
	/* Put the tag value into rid so that we can probe it ... */
	#define MPID_StartRecvTransfer( buf, size, partner, id, rid ) \
		{MPID_TRACE_CODE("BIRRRecv",id);\
		 rid = MPID_PT2PT2_TAG(id);\
		 MPID_TRACE_CODE("EIRRRecv",id);}
	#define MPID_EndRecvTransfer( buf, size, partner, id, rid ) \
		{MPID_TRACE_CODE("BIWRRecv",id);\
		 PIwrecvrr( MPID_PT2PT2_TAG(id), buf, size, MSG_OTHER, rid );\
		 MPID_TRACE_CODE("EIWRRecv",id);\
		 if (--TagsInUse == 0) CurTag = 1024; else if (id == CurTag-1) CurTag--;}
	#define MPID_TestRecvTransfer( rid ) \
		PInprobe( rid )
	#define MPID_CompleteRecvTransfer( buf, size, partner, id, rid ) \
			MPID_EndRecvTransfer( buf, size, partner, id, rid )
#endif

#ifdef PI_NO_NSEND
	#define MPID_StartSendTransfer( buf, size, partner, id, sid ) \
		{MPID_TRACE_CODE("BIRRSend",id);\
		 PIbsendrr( MPID_PT2PT2_TAG(id), buf, size, partner, MSG_OTHER );\
		 sid = 1;\
		 MPID_TRACE_CODE("EIRRSend",id);}
	#define MPID_EndSendTransfer( buf, size, partner, id, sid ) \
		{MPID_TRACE_CODE("BWRRSend",id);\
		MPID_TRACE_CODE("EWRRSend",id);}
	#define MPID_TestSendTransfer( sid ) \
		1
#else
	#define MPID_StartSendTransfer( buf, size, partner, id, sid ) \
		{MPID_TRACE_CODE("BIRRSend",id);\
		 PInsendrr( MPID_PT2PT2_TAG(id), buf, size, partner, MSG_OTHER, sid );\
		 MPID_TRACE_CODE("EIRRSend",id);}
	#define MPID_EndSendTransfer( buf, size, partner, id, sid ) \
		{MPID_TRACE_CODE("BWRRSend",id);\
		PIwsendrr( MPID_PT2PT2_TAG(id), buf, size, partner, MSG_OTHER, sid );\
		MPID_TRACE_CODE("EWRRSend",id);}
	#define MPID_TestSendTransfer( sid ) \
		PInstatus( sid )
#endif

#endif