#include <crtdbg.h>
#include "windows.h"
#include "dm.h"
#include "packets.h"
#include "queuedata.h"
#include "queue.h"

//  These are file scope
static BOOL AllocateQueueNode(PQUEUENODETYPE *pQNode);
static BOOL FreeQueueNode(PQUEUENODETYPE *pQNode);
static LONG ComparePackets(MPID_PKT_T *pPacket1,MPID_PKT_T *pPacket2);
static BOOL fQueuesInitialized = FALSE;

WINMPICHAPI QUEUETYPE Queue[2];

BOOL PacketQueuesInitialize(LONG *pnStatus)
{
	_ASSERTE(pnStatus);
	if (NULL != pnStatus)
	{
		InitializeCriticalSection(&(Queue[RECEIVEQUEUE].csQueueProtector));
		Queue[RECEIVEQUEUE].pPacketQueueHead = 0;
		Queue[RECEIVEQUEUE].pPacketQueueTail = 0;
		Queue[RECEIVEQUEUE].hSomethingQueued = CreateEvent(NULL,FALSE,FALSE,NULL);
		InitializeCriticalSection(&(Queue[SENDQUEUE].csQueueProtector));
		Queue[SENDQUEUE].pPacketQueueHead = 0;
		Queue[SENDQUEUE].pPacketQueueTail = 0;
		Queue[SENDQUEUE].hSomethingQueued = CreateEvent(NULL,FALSE,FALSE,NULL);
		if ((NULL != Queue[RECEIVEQUEUE].hSomethingQueued) &&
			(NULL != Queue[SENDQUEUE].hSomethingQueued))
			fQueuesInitialized = TRUE;
		*pnStatus = ERROR_SUCCESS;
	}
	return fQueuesInitialized;
}

BOOL Enqueue(LONG nWhichQueue,MPID_PKT_T *pIncomingPacket, LONG nSource, LONG *pnStatus)
{
	BOOL fPacketEnqueued = FALSE;

	_ASSERTE(pIncomingPacket);
	_ASSERTE(pnStatus);
	
	if ((NULL == pIncomingPacket) || (NULL == pnStatus))
		goto cleanup;
	
	*pnStatus = ERROR_SUCCESS;
	
	if (fQueuesInitialized)
	{
		PQUEUENODETYPE pTmpQNode = NULL;
		fPacketEnqueued = AllocateQueueNode(&pTmpQNode);
		if (fPacketEnqueued)
		{
			pTmpQNode->pPacket = pIncomingPacket;
			pTmpQNode->nSource = nSource;
			EnterCriticalSection(&(Queue[nWhichQueue].csQueueProtector));
			if (NULL == Queue[nWhichQueue].pPacketQueueHead)
			{
				Queue[nWhichQueue].pPacketQueueTail = pTmpQNode;
				Queue[nWhichQueue].pPacketQueueHead = pTmpQNode;
			}
			else
			{
				Queue[nWhichQueue].pPacketQueueTail->pNextQueueNode = pTmpQNode;
				Queue[nWhichQueue].pPacketQueueTail = pTmpQNode;
			}
			fPacketEnqueued = TRUE;
			fPacketEnqueued = SetEvent(Queue[nWhichQueue].hSomethingQueued);
			LeaveCriticalSection(&(Queue[nWhichQueue].csQueueProtector));
		}
		else
		{
			*pnStatus = ERROR_MALLOCFAILURE;
		}
	}
	else
	{
		*pnStatus = ERROR_QUEUESNOTINITIALIZED;
	}
cleanup:
	SetLastError(*pnStatus);
	return fPacketEnqueued;
}

BOOL Dequeue(LONG nWhichQueue,MPID_PKT_T **ppOutgoingPacket,LONG *nSource,LONG *pnStatus)
{
	BOOL fPacketDequeued = FALSE;
	PQUEUENODETYPE pTmpQNode = NULL;

	_ASSERTE(ppOutgoingPacket);
	_ASSERTE(pnStatus);

	if ((NULL == ppOutgoingPacket) || (NULL == pnStatus))
		goto cleanup;
	
	*pnStatus = ERROR_SUCCESS;
	*ppOutgoingPacket = NULL;
	*nSource = -1;
	if (fQueuesInitialized)
	{
		EnterCriticalSection(&(Queue[nWhichQueue].csQueueProtector));
		if (NULL != Queue[nWhichQueue].pPacketQueueHead)
		{
			pTmpQNode = Queue[nWhichQueue].pPacketQueueHead;
			if (Queue[nWhichQueue].pPacketQueueTail == Queue[nWhichQueue].pPacketQueueHead)
			{
				Queue[nWhichQueue].pPacketQueueTail = NULL;
				ResetEvent(Queue[nWhichQueue].hSomethingQueued);
			}
			else
				SetEvent(Queue[nWhichQueue].hSomethingQueued);

			Queue[nWhichQueue].pPacketQueueHead = pTmpQNode->pNextQueueNode;
			fPacketDequeued = TRUE;
		}
		else
		{
			*pnStatus = ERROR_QUEUEEMPTY;
		}
		LeaveCriticalSection(&(Queue[nWhichQueue].csQueueProtector));
		if (fPacketDequeued)
		{
			*ppOutgoingPacket = pTmpQNode->pPacket;
			*nSource = pTmpQNode->nSource;
			FreeQueueNode(&pTmpQNode);
		}
	}
	else
	{
		*pnStatus = ERROR_QUEUESNOTINITIALIZED;
	}
cleanup:
	SetLastError(*pnStatus);
	return fPacketDequeued;
}

BOOL IsEmptyQueue(LONG nWhichQueue)
{
	BOOL fEmpty = TRUE;

	EnterCriticalSection(&(Queue[nWhichQueue].csQueueProtector));
	if (NULL != Queue[nWhichQueue].pPacketQueueHead)
		fEmpty = FALSE;
	LeaveCriticalSection(&(Queue[nWhichQueue].csQueueProtector));

	return fEmpty;
}

BOOL ResetPacketEvent(LONG nWhichQueue)
{
	BOOL fEmpty = TRUE;

	EnterCriticalSection(&(Queue[nWhichQueue].csQueueProtector));
	if (NULL != Queue[nWhichQueue].pPacketQueueHead)
		fEmpty = FALSE;
	if (fEmpty)
		ResetEvent(Queue[nWhichQueue].hSomethingQueued);
	LeaveCriticalSection(&(Queue[nWhichQueue].csQueueProtector));
	return fEmpty;
}

static BOOL AllocateQueueNode(PPQUEUENODETYPE ppQNode)
{
	BOOL fAllocationSuccessful = TRUE;

	_ASSERTE(ppQNode);
	if (NULL != ppQNode)
	{
		*ppQNode = NULL;
		*ppQNode = (QUEUENODETYPE*)malloc(sizeof(QUEUENODETYPE));
		_ASSERTE(*ppQNode);
		if (NULL != *ppQNode)
		{
			(*ppQNode)->pNextQueueNode = NULL;
			(*ppQNode)->pPacket = NULL;
		}
		else
		{
			fAllocationSuccessful = FALSE;
		}
	}
	else
	{
		fAllocationSuccessful = FALSE;
	}
	return fAllocationSuccessful;
}

static LONG ComparePackets(MPID_PKT_T *pPacket1,MPID_PKT_T *pPacket2)
{
	LONG nResult = 0;
	LONG size = 0;
	// for now, compare the pointers, if the same, then same packet
	_ASSERTE(pPacket1);
	_ASSERTE(pPacket2);
	size = sizeof(MPID_PKT_HEAD_T) + pPacket1->head.len;
	nResult = memcmp(pPacket1,pPacket2,size);
	return nResult;
}


DWORD AllocateQueuePacket(MPID_PKT_T **pPkt)
{
	DWORD dwStatus = !ERROR_SUCCESS;

	_ASSERTE(pPkt);
	if (pPkt)
	{
		*pPkt = (MPID_PKT_T*)malloc(sizeof(MPID_PKT_T));
		_ASSERT(*pPkt);
		if (*pPkt)
			dwStatus = ERROR_SUCCESS;
		else
			dwStatus = GetLastError();
	}
	return dwStatus;
}

DWORD AllocateQueueBuffer(BYTE **pBuf,DWORD dwSize)
{
	DWORD dwStatus = !ERROR_SUCCESS;

	_ASSERTE(pBuf);
	_ASSERTE(0 <dwSize);
	if (pBuf && dwSize)
	{
		*pBuf = (BYTE*)malloc(dwSize);
		_ASSERTE(*pBuf);
		if(*pBuf)
			dwStatus = ERROR_SUCCESS;
		else
			dwStatus = GetLastError();
	}
	return dwStatus;
}

void FreeQueuePacket(MPID_PKT_T *pPkt)
{
	_ASSERTE(pPkt);
	if (pPkt)
		free(pPkt);
}

void FreeQueueBuffer(BYTE *pBuf)
{
	_ASSERTE(pBuf);
	if (pBuf)
		free(pBuf);
}

static BOOL FreeQueueNode(PPQUEUENODETYPE ppQNode)
{
	BOOL fFreeSuccessful = TRUE;

	_ASSERTE(ppQNode);
	_ASSERTE(*ppQNode);
	if ((NULL != ppQNode) && (NULL != *ppQNode))
	{
		(*ppQNode)->pNextQueueNode = NULL;
		(*ppQNode)->pPacket = NULL;
		free(*ppQNode);
		*ppQNode = NULL;
	}
	else
	{
		fFreeSuccessful = FALSE;
	}
	return fFreeSuccessful;
}
