#include <crtdbg.h>
#include "windows.h"
#include "dm.h"
#include "packets.h"
#include "queue.h"
#include "queuedata.h"
#include "stdio.h"
#include "tchar.h"
#include "processinfo.h"
#include "infoprint.h"

DWORD WINAPI SHMEMSendThread(LPVOID lpV)
{
	extern int MPID_MyWorldRank;
	extern int MPID_WorldSize;
	extern DWORD dwMyMachineID;
	extern DWORD dwNumLocalProcesses;
	extern BOOL TCPIPProcNeeded;
	extern MPID_WINNT_SHMEM_BUFFER_T *ShmemBuffer[MAX_NUM_SHMEM_PROCS];
	extern MPID_WINNT_SHMEM_BUFFER_T *TCPIPShmemBuffer;
	extern QUEUETYPE Queue[2];
	extern HANDLE hEndProcessEvent;
	extern JOBINFO *JobInfo;
	extern WORLDPROCESSMAP *WorldProcessMap;
	DWORD dwStatus = ERROR_SUCCESS;
	HANDLE hEvent[2];
	MPID_PKT_T *pTmpPkt = NULL;
	MPID_PKT_T *pShmemPkt = NULL;
	DWORD dwSource;
	DWORD dwDestination;
	DWORD dwBytesToSend;
	BOOL fStatus = FALSE;
	BYTE *pCopyFrom = NULL;
	DWORD dwBytesSent = 0;
	MPID_WINNT_SHMEM_BUFFER_T *pShmemBuffer = NULL;

#ifdef _DEBUG
	DWORD i;
	for (i=0; i < dwNumLocalProcesses; i++)
		_ASSERTE(ShmemBuffer[i]);
#endif
	_ASSERTE(WorldProcessMap);

	hEvent[0] = hEndProcessEvent;
	hEvent[1] = Queue[SENDQUEUE].hSomethingQueued;

	while (TRUE)
	{
		dwStatus = WaitForMultipleObjects(2,hEvent,FALSE,INFINITE);
		_ASSERTE(WAIT_FAILED != dwStatus);
		if (WAIT_FAILED == dwStatus)
			goto reporterror;
		if (WAIT_OBJECT_0 == dwStatus)
			goto cleanup;
		Dequeue(SENDQUEUE,&pTmpPkt,&dwSource,&dwStatus);
		dwDestination = pTmpPkt->head.dest;
		_ASSERTE((0 <= dwDestination) && (dwDestination < WorldProcessMap->dwNumProcs));
		if (dwDestination == (DWORD)MPID_MyWorldRank)
		{
			Enqueue(RECEIVEQUEUE,pTmpPkt,dwSource,&dwStatus);
			continue;
		}
		else
		{
			switch (WorldProcessMap->Process[dwDestination].eProtocol)
			{
			case E_SHMEM_PROTOCOL:
				pShmemBuffer = ShmemBuffer[WorldProcessMap->Process[dwDestination].dwLocalBufferNumber];
				break;
			case E_TCPIP_PROTOCOL:
				_ASSERTE(TCPIPShmemBuffer);
				pShmemBuffer = TCPIPShmemBuffer;
				break;
			default:
				_ASSERTE(FALSE);
				/* We have no idea what protocol to use. */
				break;
			}
			pShmemPkt = &(pShmemBuffer->Packet);
		}
		dwStatus = WaitForSingleObject(pShmemBuffer->hBufferEmpty,INFINITE);
		_ASSERTE(WAIT_OBJECT_0 == dwStatus);
		if (WAIT_FAILED == dwStatus)
			goto reporterror;
		dwStatus = WaitForSingleObject(pShmemBuffer->hMutex,INFINITE);
		_ASSERTE(WAIT_OBJECT_0 == dwStatus);
		if (WAIT_FAILED == dwStatus)
			goto reporterror;
		switch (pTmpPkt->head.mode)
		{
		case MPID_PKT_SHORT:
			dwBytesToSend =   sizeof(MPID_PKT_HEAD_T)
							+ pTmpPkt->head.len;
			memcpy(pShmemPkt,pTmpPkt,dwBytesToSend);
			break;
		case MPID_PKT_SHORT_SYNC:
			dwBytesToSend =   sizeof(MPID_PKT_HEAD_T)
							+ sizeof(DECL_SYNC_ID_T)
							+ pTmpPkt->head.len;
			memcpy(pShmemPkt,pTmpPkt,dwBytesToSend);
			break;
		case MPID_PKT_LONG:
			dwBytesToSend = sizeof(MPID_PKT_LONG_T);
			memcpy(pShmemPkt,pTmpPkt,dwBytesToSend);
			pCopyFrom = pTmpPkt->long_pkt.pBuffer;
			dwBytesSent = 0;
			while (dwBytesSent < (DWORD)pTmpPkt->head.len)
			{
				fStatus = SetEvent(pShmemBuffer->hBufferFull);
				_ASSERTE(fStatus);
				fStatus = ReleaseMutex(pShmemBuffer->hMutex);
				_ASSERTE(fStatus);
				dwStatus = WaitForSingleObject(pShmemBuffer->hBufferEmpty,INFINITE);
				_ASSERTE(WAIT_OBJECT_0 == dwStatus);
				if (WAIT_FAILED == dwStatus)
					goto reporterror;
				dwStatus = WaitForSingleObject(pShmemBuffer->hMutex,INFINITE);
				_ASSERTE(WAIT_OBJECT_0 == dwStatus);
				if (WAIT_FAILED == dwStatus)
					goto reporterror;
				if (MPID_PKT_MAX_DATA_SIZE < (pTmpPkt->head.len - dwBytesSent))
					dwBytesToSend = MPID_PKT_MAX_DATA_SIZE;
				else
					dwBytesToSend = pTmpPkt->head.len - dwBytesSent;
				memcpy(pShmemPkt,pTmpPkt,sizeof(MPID_PKT_HEAD_T));
				pShmemPkt->head.mode = MPID_PKT_LONG_DATA;
				pShmemPkt->head.len = dwBytesToSend;
				memcpy(pShmemPkt->long_data_pkt.buffer,pCopyFrom,dwBytesToSend);
				pCopyFrom += dwBytesToSend;
				dwBytesSent += dwBytesToSend;
			}
			free(pTmpPkt->long_pkt.pBuffer);
			break;
		case MPID_PKT_LONG_SYNC:
			dwBytesToSend = sizeof(MPID_PKT_LONG_SYNC_T);
			memcpy(pShmemPkt,pTmpPkt,dwBytesToSend);
			pCopyFrom = pTmpPkt->long_sync_pkt.pBuffer;
			dwBytesSent = 0;
			while (dwBytesSent < (DWORD)pTmpPkt->head.len)
			{
				fStatus = SetEvent(pShmemBuffer->hBufferFull);
				_ASSERTE(fStatus);
				fStatus = ReleaseMutex(pShmemBuffer->hMutex);
				_ASSERTE(fStatus);
				dwStatus = WaitForSingleObject(pShmemBuffer->hBufferEmpty,INFINITE);
				_ASSERTE(WAIT_FAILED != dwStatus);
				if (WAIT_FAILED == dwStatus)
					goto reporterror;
				dwStatus = WaitForSingleObject(pShmemBuffer->hMutex,INFINITE);
				_ASSERTE(WAIT_FAILED != dwStatus);
				if (WAIT_FAILED == dwStatus)
					goto reporterror;
				if (MPID_PKT_MAX_DATA_SIZE < (pTmpPkt->head.len - dwBytesSent))
					dwBytesToSend = MPID_PKT_MAX_DATA_SIZE;
				else
					dwBytesToSend = pTmpPkt->head.len - dwBytesSent;
				memcpy(pShmemPkt,pTmpPkt,sizeof(MPID_PKT_HEAD_T));
				pShmemPkt->head.mode = MPID_PKT_LONG_DATA;
				pShmemPkt->head.len = dwBytesToSend;
				memcpy(pShmemPkt->long_data_pkt.buffer,pCopyFrom,dwBytesToSend);
				pCopyFrom += dwBytesToSend;
				dwBytesSent += dwBytesToSend;
			}
			free(pTmpPkt->long_sync_pkt.pBuffer);
			break;
		case MPID_PKT_SYNC_ACK:
			dwBytesToSend = sizeof(MPID_PKT_SYNC_ACK_T);
			memcpy(pShmemPkt,pTmpPkt,dwBytesToSend);
			break;
		}
		fStatus = SetEvent(pShmemBuffer->hBufferFull);
		_ASSERTE(fStatus);
		fStatus = ReleaseMutex(pShmemBuffer->hMutex);
		_ASSERTE(fStatus);
		free(pTmpPkt);
	}

cleanup:
	_ASSERTE(ERROR_SUCCESS == dwStatus);
	return dwStatus;
reporterror:
	dwStatus = GetLastError();
	goto cleanup;
}