/*
 *  $Id: chinit.c,v 1.39 1996/01/11 18:34:40 gropp Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

#ifndef lint
static char vcid[] = "$Id: chinit.c,v 1.39 1996/01/11 18:34:40 gropp Exp $";
#endif

int MPID_WINNT_Complete_pending();
int MPID_GetMsgDebugFlag();

/* 
    This file contains the routines that provide the basic information 
    on the device, and initialize it
 */

#if defined(rs6000) && !defined(_ALL_SOURCE)
/* AIX's version of netinet/in.h REQUIRES but DOES NOT CHECK 
   for _ALL_SOURCE! 
   (This is needed below incase sys/types.h is loaded in these includes)
 */
#define _ALL_SOURCE
#endif

#include "mpid.h"
/*#include "system/system.h"*/
/* #CMMD DECLARATION# */
#include "time.h"
#include "windows.h"
#include <crtdbg.h>

void (*MPID_ErrorHandler)() = MPID_DefaultErrorHandler;
/* For tracing channel operations by ADI underlayer */
FILE *MPID_TRACE_FILE = 0;

/* For debugging statements */
FILE *MPID_DEBUG_FILE = 0;

#ifdef MPID_PKT_VAR_SIZE
int MPID_PKT_DATA_SIZE = MPID_PKT_MAX_DATA_SIZE;
int MPID_SetPktSize( len )
int len;
{
if (len < 0) return MPID_PKT_MAX_DATA_SIZE;
if (len > MPID_PKT_MAX_DATA_SIZE) len = MPID_PKT_MAX_DATA_SIZE;
MPID_PKT_DATA_SIZE = len;
return len;
}
#else
int MPID_SetPktSize( len )
int len;
{
return MPID_PKT_MAX_DATA_SIZE;
}
#endif

/* #define DEBUG(a) {a} */
#define DEBUG(a)

static int DebugSpace = 0;
int MPID_DebugFlag = 0;

void MPID_SetSpaceDebugFlag( flag )
int flag;
{
DebugSpace = flag;
#ifdef CHAMELEON_COMM   /* #CHAMELEON_START# */
/* This file may be used to generate non-Chameleon versions */
if (flag) {
    /* Check the validity of the malloc arena on every use of trmalloc/free */
    trDebugLevel( 1 );
    }
#endif                  /* #CHAMELEON_END# */
}
void MPID_SetDebugFlag( ctx, f )
void *ctx;
int f;
{
MPID_DebugFlag = f;
}
void MPID_SetDebugFile( name )
char *name;
{
char filename[1024];

if (strcmp( name, "-" ) == 0) {
    MPID_DEBUG_FILE = stdout;
    return;
    }
if (strchr( name, '%' )) {
    sprintf( filename, name, MPID_MyWorldRank );
    MPID_DEBUG_FILE = fopen( filename, "w" );
    }
else
    MPID_DEBUG_FILE = fopen( name, "w" );

if (!MPID_DEBUG_FILE) MPID_DEBUG_FILE = stdout;
}
void MPID_Set_tracefile( name )
char *name;
{
char filename[1024];

if (strcmp( name, "-" ) == 0) {
    MPID_TRACE_FILE = stdout;
    return;
    }
if (strchr( name, '%' )) {
    sprintf( filename, name, MPID_MyWorldRank );
    MPID_TRACE_FILE = fopen( filename, "w" );
    }
else
    MPID_TRACE_FILE = fopen( name, "w" );

/* Is this the correct thing to do? */
if (!MPID_TRACE_FILE)
    MPID_TRACE_FILE = stdout;
}


#ifndef MPID_STAT_NONE
int MPID_n_short       = 0,         /* short messages */
    MPID_n_long        = 0,         /* long messages */
    MPID_n_unexpected  = 0,         /* unexpected messages */
    MPID_n_syncack     = 0;         /* Syncronization acknowledgments */
#endif

/***************************************************************************/
/* Some operations are completed in several stages.  To ensure that a      */
/* process does not exit from MPID_End while requests are pending, we keep */
/* track of how many are out-standing                                      */
/***************************************************************************/
int MPID_n_pending     = 0;         /* Number of uncompleted split requests */

/*****************************************************************************
  Here begin the interface routines themselves
 *****************************************************************************/
void MPID_WINNT_Myrank( rank )
int *rank;
{
*rank = MPID_MyWorldRank;
}

void MPID_WINNT_Mysize( size )
int *size;
{
*size = MPID_WorldSize;
}

/* 
    In addition, Chameleon processes many command-line arguments 

    This should return a structure that contains any relavent context
    (for use in the multiprotocol version)

    This version currently returns null, as all data is static.
 */
void *MPID_WINNT_Init( argc, argv )
int  *argc;
char ***argv;
{
/* Set the file for Debugging output.  The actual output is controlled
   by MPIDDebugFlag */
if (MPID_DEBUG_FILE == 0) MPID_DEBUG_FILE = stdout;

/*PIiInit( argc, argv );*/
MPID_WINNT_init (argc, argv);
#ifdef MPID_DEBUG_ALL   /* #DEBUG_START# */
DEBUG(fprintf(MPID_DEBUG_FILE,"[%d] Finished init\n", MPID_MyWorldRank );)
#endif                  /* #DEBUG_END# */

/* Turn off the resource monitors */
#if !defined(euih) && !defined(eui)
/* If we are euih or SP2 eui, we can't use SIGALRM; this call sets SIGALRM to 
   SIG_IGN */
/*SYSetResourceClockOff();*/
#endif


/* Initialize any data structures in the send and receive handlers */
MPID_WINNT_Init_recv_code();
MPID_WINNT_Init_send_code();

#ifdef MPID_DEBUG_ALL   /* #DEBUG_START# */
DEBUG(fprintf(MPID_DEBUG_FILE,"[%d] leaving chinit\n", MPID_MyWorldRank );)
#endif                  /* #DEBUG_END# */

return (void *)0;
}

/* Barry Smith suggests that this indicate who is aborting the program.
   There should probably be a separate argument for whether it is a 
   user requested or internal abort.
 */
void MPID_WINNT_Abort( code )
int code;
{
fprintf( stderr, "[%d] Aborting program; exit code [%d].\n", MPID_MyWorldRank, code );
fflush( stderr );
fflush( stdout );
/*SYexitall( (char *)0, code );*/
_ASSERTE(FALSE);
MPID_WINNT_abort();
}

void MPID_WINNT_End()
{
#ifdef MPID_DEBUG_ALL   /* #DEBUG_START# */
if (MPID_DebugFlag) {
    fprintf( MPID_DEBUG_FILE,
   "[%d] Entering MPID_End\n", MPID_MyWorldRank );
    }
#endif                  /* #DEBUG_END# */
/* Finish off any pending transactions */
MPID_WINNT_Complete_pending();

if (MPID_GetMsgDebugFlag()) {
    MPID_PrintMsgDebug();
    }
#ifdef CHAMELEON_COMM       /* #CHAMELEON_START# */
if (DebugSpace)
    trdump( stdout );
#endif                      /* #CHAMELEON_END# */
/* We should really generate an error or warning message if there 
   are uncompleted operations... */
/*PIiFinish();*/
MPID_WINNT_finalize();
}

#include "processinfo.h"

void MPID_WINNT_Node_name( name, len )
char *name;
int  len;
{
	extern JOBINFO *JobInfo;
	extern int dwMyMachineID;

	/*PINodeName( name, len );*/
	char myname[256];
	sprintf(myname,"%s:[%d]",JobInfo->Machine[dwMyMachineID].szMachineName,MPID_MyWorldRank);
	strncpy(name,myname,len);
}

void MPID_WINNT_Version_name( name )
char *name;
{
sprintf( name, "ADI version %4.2f - transport %s", MPIDPATCHLEVEL, 
	 MPIDTRANSPORT );
}

#if defined(HAVE_GETTIMEOFDAY) || defined(USE_WIERDGETTIMEOFDAY)
#include <sys/types.h>
#include <time.h>
#endif
/* I don't know what the correct includes are for the other versions... */


#define FACTOR   4294967296	/* equivalint to shift 32 positions left */

static double dSecInTick = -1.0;

double MPID_WINNT_Wtime()
{
	static double dSecInTick = -1.0;
	double dTime;
	LARGE_INTEGER CurTime; 
	BOOL bStatus;

	if (dSecInTick < 0) 
	{
		LARGE_INTEGER PerfFreq;
		bStatus = QueryPerformanceFrequency( &PerfFreq );
		if (0 == bStatus)
		{
			/*The system does not have high-performance 
			counter installed */
			return 0.0;
		}
		dSecInTick = 1.0/((double)PerfFreq.HighPart*(double)FACTOR+(double)PerfFreq.LowPart);
	}		

	bStatus = QueryPerformanceCounter( &CurTime );
	if (!bStatus)
		return 0.0;
	
	dTime =   ((double)CurTime.HighPart * (double)FACTOR)
			+ ((double)CurTime.LowPart);

	dTime *= dSecInTick;

	return dTime;
}


/* This returns a value that is correct but not the best value that
   could be returned.
   It makes several separate stabs at computing the tickvalue.
*/
double MPID_WINNT_Wtick()
{
	LARGE_INTEGER PerfFreq;
	BOOL bStatus;

	if (dSecInTick < 0) 
	{
		LARGE_INTEGER PerfFreq;
		bStatus = QueryPerformanceFrequency( &PerfFreq );
		if (0 == bStatus)
		{
			/*The system does not have high-performance 
			counter installed */
			return 1.0;
		}
		dSecInTick = 1.0/((double)PerfFreq.HighPart*(double)FACTOR+(double)PerfFreq.LowPart);
	}
	return dSecInTick;
}


void MPID_WINNT_Error_handler( r )
void (*r)();
{
if (r)
    MPID_ErrorHandler = r;
else
    MPID_ErrorHandler = MPID_DefaultErrorHandler;
}

/* This is the "panic" handler.  Correctable errors should be passed on
   to the user (see MPID_WINNTK_MSGLEN) */
void MPID_DefaultErrorHandler( code, str )
int  code;
char *str;
{
if (str) 
    fprintf( stderr, "[%d] %s\n", MPID_MyWorldRank, str );
MPID_WINNT_Abort( code );
}


/* We also need an "ErrorsReturn" and a sensible error return strategy */

/*
   Data about messages
 */
static int DebugMsgFlag = 0;
void MPID_SetMsgDebugFlag( f )
int f;
{
DebugMsgFlag = f;
}
int MPID_GetMsgDebugFlag()
{
return DebugMsgFlag;
}
void MPID_PrintMsgDebug()
{
#ifndef MPID_STAT_NONE
fprintf( stdout, "[%d] short = %d, long = %d, unexpected = %d, ack = %d\n",
	 MPID_MyWorldRank, MPID_n_short, MPID_n_long, MPID_n_unexpected, 
	 MPID_n_syncack );
#endif
}
