#include <stdio.h>
#include <windows.h>
#include <windowsx.h>
#include <crtdbg.h>
#include "packets.h"
#include "queue.h"
#include "processinfo.h"
#include "infoprint.h"

HANDLE hWorldProcessMapFileMap = (HANDLE)0;
HANDLE hJobInfoFileMap = (HANDLE)0;
HANDLE hLocalProcessHandles = (HANDLE)0;
HANDLE hEndProcessEvent = (HANDLE)0;
HANDLE hTCPIPShmemBuf = (HANDLE)0;
HANDLE hWorkerThreads[3];

WINMPICHAPI JOBINFO *JobInfo = NULL;
WINMPICHAPI WORLDPROCESSMAP *WorldProcessMap = NULL;
WINMPICHAPI LOCALPROCESSHANDLES *LocalProcessHandles = NULL;

int MPID_MyWorldRank = 0;
int MPID_WorldSize = 0;
DWORD dwMyMachineID = 0;
DWORD dwNumLocalProcesses = 0;
DWORD dwNumMachines = 0;
BOOL TCPIPProcNeeded = FALSE;
BOOL IamTCPIPProc = FALSE;
MPID_WINNT_SHMEM_BUFFER_T *ShmemBuffer[MAX_NUM_SHMEM_PROCS];
MPID_WINNT_SHMEM_BUFFER_T *TCPIPShmemBuffer = NULL;


void* MPID_WINNT_init (int* argc, char*** argv)
{
	DWORD dwStatus = ERROR_SUCCESS;
	BOOL fInitialized = FALSE;
	BOOL fStatus = FALSE;
	BOOL fFound = FALSE;
	DWORD i;
	SECURITY_ATTRIBUTES sa;
	DWORD dwThreadID;
	MPID_WINNT_SHMEM_BUFFER_T *pWhichBuffer = NULL;

	if (7 > *argc)
	{
		printf("The necessary command line information to run this MPI process\n");
		printf("is not present.\n");
		dwStatus = (DWORD)-999;
		goto reporterror;
	}

	for (i=0; i < MAX_NUM_SHMEM_PROCS; i++)
		ShmemBuffer[i] = NULL;

	hWorkerThreads[0] = NULL;
	hWorkerThreads[1] = NULL;
	hWorkerThreads[2] = NULL;

	for (i=1; i < (DWORD)*argc; i++)
	{
		if (!strcmp(CMD_TCPIPPROC_NOTIFY_STR,(*argv)[i]))
		{
			IamTCPIPProc = TRUE;
			break;
		}
	}
	for (i=1; i < (DWORD)*argc; i++)
	{
		if (!strcmp(CMD_TCPIPPROC_STR,(*argv)[i]))
		{
			TCPIPProcNeeded = TRUE;
			break;
		}
	}
	fFound = FALSE;
	for (i=1; i < (DWORD)*argc; i++)
	{
		if (!strcmp(CMD_GLOBALRANK_STR,(*argv)[i]))
		{
			MPID_MyWorldRank = (DWORD)atol((*argv)[i+1]);
			fInitialized = TRUE;
			fFound = TRUE;
			break;
		}
	}
	if (!fFound)
	{
		printf("No [%s] command line argument!\n",CMD_GLOBALRANK_STR);
		goto cleanup;
	}

	fFound = FALSE;
	for (i=1; i < (DWORD)*argc; i++)
	{	
		if (!strcmp(CMD_JOBINFO_STR,(*argv)[i]))
		{
			hJobInfoFileMap = (HANDLE)atol((*argv)[i+1]);
			_ASSERTE(hJobInfoFileMap);
			fInitialized = fInitialized && TRUE;
			fFound = TRUE;
			break;
		}
	}
	if (!fFound)
	{
		printf("No [%s] command line argument!\n",CMD_JOBINFO_STR);
		goto cleanup;
	}

	fFound = FALSE;
	for (i=1; i < (DWORD)*argc; i++)
	{
		if (!strcmp(CMD_WORLDMAP_STR,(*argv)[i]))
		{
			hWorldProcessMapFileMap = (HANDLE)atol((*argv)[i+1]);
			_ASSERTE(hWorldProcessMapFileMap);
			fInitialized = fInitialized && TRUE;
			fFound = TRUE;
			break;
		}
	}
	if (!fFound)
	{
		printf("No [%s] command line argument!\n",CMD_WORLDMAP_STR);
		goto cleanup;
	}

	_ASSERTE(fInitialized);
	if (!fInitialized)
	{
		dwStatus = (DWORD)-998;
		goto reporterror;
	}

	WorldProcessMap = MapViewOfFile(hWorldProcessMapFileMap,
									FILE_MAP_ALL_ACCESS,
									0,
									0,
									0);
	_ASSERTE(WorldProcessMap);
	if (NULL == WorldProcessMap)
		goto geterror;
	JobInfo = MapViewOfFile(hJobInfoFileMap,
							FILE_MAP_ALL_ACCESS,
							0,
							0,
							0);
	_ASSERTE(JobInfo);
	if (NULL == JobInfo)
		goto geterror;

	MPID_WorldSize = JobInfo->dwWorldSize;
	dwMyMachineID = JobInfo->dwMyMachineID;
	dwNumMachines = JobInfo->dwNumMachines;
	dwNumLocalProcesses = JobInfo->Machine[dwMyMachineID].dwNumProcesses;


	_ASSERTE(JobInfo->hLocalProcessHandles);
	hLocalProcessHandles = JobInfo->hLocalProcessHandles;
	LocalProcessHandles = MapViewOfFile(hLocalProcessHandles,
										FILE_MAP_ALL_ACCESS,
										0,
										0,
										0);
	_ASSERTE(LocalProcessHandles);
	if (NULL == LocalProcessHandles)
		goto geterror;


	sa.nLength = sizeof(SECURITY_ATTRIBUTES);
	sa.lpSecurityDescriptor = NULL;
	sa.bInheritHandle = TRUE;

	hEndProcessEvent = CreateEvent(&sa,
								   TRUE,
								   FALSE,
								   NULL);
	_ASSERTE(hEndProcessEvent);
	if (NULL == hEndProcessEvent)
		goto geterror;

	for (i=0; i < dwNumLocalProcesses; i++)
	{
		_ASSERTE(JobInfo->Machine[dwMyMachineID].Process[i].hSharedMemoryBuffer);
		ShmemBuffer[i] = (MPID_WINNT_SHMEM_BUFFER_T *)MapViewOfFile(JobInfo->Machine[dwMyMachineID].Process[i].hSharedMemoryBuffer,
																	FILE_MAP_ALL_ACCESS,
																	0,
																	0,
																	0);
		_ASSERTE(ShmemBuffer[i]);
		if (NULL == ShmemBuffer[i])
			goto geterror;
	}
	if (TCPIPProcNeeded)
	{
		_ASSERTE(JobInfo->hTCPIPProcBuffer);
		hTCPIPShmemBuf = JobInfo->hTCPIPProcBuffer;
		TCPIPShmemBuffer = MapViewOfFile(hTCPIPShmemBuf,
										FILE_MAP_ALL_ACCESS,
										0,
										0,
										0);
		_ASSERTE(TCPIPShmemBuffer);
		if (NULL == TCPIPShmemBuffer)
			goto geterror;
	}

	fStatus = PacketQueuesInitialize(&dwStatus);
	_ASSERTE(fStatus);

	/* I should now have everything I need to run the MPI process */
	hWorkerThreads[0] = CreateThread(&sa,
									 0,
									 SHMEMSendThread,
									 NULL,
									 0,
									 &dwThreadID);
	_ASSERTE(hWorkerThreads[0]);
	if (NULL == hWorkerThreads[0])
		goto geterror;

	if (!IamTCPIPProc)
	{
		pWhichBuffer = ShmemBuffer[WorldProcessMap->Process[MPID_MyWorldRank].dwLocalBufferNumber];
	}
	else
	{
		_ASSERTE(TCPIPShmemBuffer);
		pWhichBuffer = TCPIPShmemBuffer;
	}

	hWorkerThreads[1] = CreateThread(&sa,
									 0,
									 SHMEMReceiveThread,
									 pWhichBuffer,
									 0,
									 &dwThreadID);
	_ASSERTE(hWorkerThreads[1]);
	if (NULL == hWorkerThreads[1])
		goto geterror;

	hWorkerThreads[2] = CreateThread(&sa,
									 0,
									 GLOBALTerminatorThread,
									 NULL,
									 0,
									 &dwThreadID);
	_ASSERTE(hWorkerThreads[2]);
	if (NULL == hWorkerThreads[2])
		goto geterror;

	Sleep(50); /* Give all the threads a chance to get started */

	MPID_WINNT_Barrier();

cleanup:
	return (void *)dwStatus;
geterror:
	dwStatus = GetLastError();
reporterror:
	printf("Failure in initilization function MPID_WINNT_init\n");
	printf("Unable to continue MPI process.\n");
	Sleep(1000);
	ExitProcess(dwStatus);
	goto cleanup;
}

void* MPID_WINNT_WorldSize(int* size)
{
	*size = MPID_WorldSize;
	return 0;
}

void* MPID_WINNT_MyRank(int* rank)
{
	*rank = MPID_MyWorldRank;
	return 0;
}

void MPID_WINNT_CleanUp()
{
	DWORD i;

	for (i=0;i < MAX_NUM_SHMEM_PROCS; i++)
	{
		if (ShmemBuffer[i])
			UnmapViewOfFile(ShmemBuffer[i]);
	}
	if (JobInfo)
		UnmapViewOfFile(JobInfo);
	if (WorldProcessMap)
		UnmapViewOfFile(WorldProcessMap);
}

void MPID_WINNT_finalize()
{
	DWORD dwStatus;
	BOOL fRetval;

	while(!IsEmptyQueue(SENDQUEUE))
		Sleep(50);
	
	fRetval = SetEvent(hEndProcessEvent);
	_ASSERTE(fRetval);

	dwStatus = 	WaitForMultipleObjects(3, 
									   hWorkerThreads,
									   TRUE,
									   INFINITE);
	_ASSERTE(WAIT_FAILED != dwStatus);

	MPID_WINNT_CleanUp();

	if (hWorkerThreads[0])
	{
		fRetval = CloseHandle(hWorkerThreads[0]);
		_ASSERTE(fRetval);
	}
	if (hWorkerThreads[1])
	{
		fRetval = CloseHandle(hWorkerThreads[1]);
		_ASSERTE(fRetval);
	}
	if (hWorkerThreads[2])
	{
		fRetval = CloseHandle(hWorkerThreads[2]);
		_ASSERTE(fRetval);
	}
	fRetval = CloseHandle(hEndProcessEvent);
	_ASSERTE(fRetval);

	fflush(stdout);
	fflush(stderr);
	exit(0);
}

void MPID_WINNT_Barrier()
{
	DWORD dwStatus = ERROR_SUCCESS;
	BOOL fStatus = FALSE;

	/*
		If I am the only process (at all) or if I am
		the TCPIP agent, there is no need
		to block at the barrier because there is no one to
		synchronize with me
	*/
	if (IamTCPIPProc || (2 > dwNumLocalProcesses))
		return;
	if (1 == dwNumMachines)
	{
		/*
			don't wait here because if I fail, I know that I am the last
			to get here
		*/
		dwStatus = WaitForSingleObject(JobInfo->hBarrierSemaphore, 0);

		_ASSERTE(WAIT_FAILED != dwStatus);

		if (WAIT_OBJECT_0 == dwStatus)
		{
			dwStatus = WaitForSingleObject(JobInfo->hBarrierEvent, INFINITE);
			_ASSERTE(WAIT_FAILED != dwStatus);
		}
		else /* Must be a WAIT_TIMEOUT then */
		{
			fStatus = ReleaseSemaphore(JobInfo->hBarrierSemaphore,
									   dwNumLocalProcesses - 1,
									   NULL);
			_ASSERTE(fStatus);
			fStatus = PulseEvent(JobInfo->hBarrierEvent);
			_ASSERTE(fStatus); 
		}
	}
	else
	{
	/* TODO:  Have to do something else with the network */
	}
}

void MPID_WINNT_abort()
{
	BOOL fStatus;

	MPID_WINNT_CleanUp();
	fStatus = SetEvent(JobInfo->hGlobalTerminatorEvent);
	_ASSERTE(fStatus);
	Sleep(INFINITE);

}

DWORD WINAPI GLOBALTerminatorThread (LPVOID lpV)
{
	char msg[128];
	HANDLE hEvents[2]; 
	DWORD dwWhichEvent;
	hEvents[0] = hEndProcessEvent;
	hEvents[1] = JobInfo->hGlobalTerminatorEvent;


	dwWhichEvent = WaitForMultipleObjects(2,
										  hEvents, 
										  FALSE,
										  INFINITE);

	_ASSERTE(WAIT_FAILED != dwWhichEvent);
		
	if (WAIT_OBJECT_0 == dwWhichEvent)
	{
		OutputDebugString("Normal exit of the GLOBALTerminator thread");
		ExitThread(0);
	}
	else
	{
		sprintf (msg, "Process [%d]: the GLOBALTerminator thread got termination signal.Aborting the process.", MPID_MyWorldRank);
		OutputDebugString(msg);
		ExitProcess(0xFFFFFFFF);
	}

	return 0xFFFFFFFF;
}
