#include "windows.h"
#include "packets.h"
#include "processinfo.h"
#include "infoprint.h"
#include "mpirun.h"
#include "debuglog.h"
#include "crtdbg.h"

extern BOOL fDebug;
__declspec(thread) PROCESS_INFORMATION piTCP;


void WaitForTCPToFinish()
{
	DWORD dwStatus;
	//_ASSERTE(piTCP.hProcess);
	if (piTCP.hProcess)
	{
		dwStatus = WaitForSingleObject(piTCP.hProcess,INFINITE);
		_ASSERTE(WAIT_FAILED != dwStatus);
	}
}


DWORD RunLocalMPIJob(JOBINFO *JobInfo,
					 HANDLE hJobInfoFileMap,
					 WORLDPROCESSMAP *WorldProcessMap,
					 HANDLE hWorldProcessMapFileMap,
					 LOCALPROCESSHANDLES *LocalProcessHandles,
					 HANDLE hLocalProcessHandles, HANDLE m_hFile)
{
	DWORD dwStatus = ERROR_SUCCESS;
	SECURITY_ATTRIBUTES sa;
	STARTUPINFO si;
	PROCESS_INFORMATION piProcess[MAX_NUM_SHMEM_PROCS];
	DWORD dwBufSize = 0;
	MPID_WINNT_SHMEM_BUFFER_T *pBuf;
	DWORD i;
	char szMsg[256];
	BOOL fStatus = TRUE;
	char szArgs[MAXPATHLEN];

	memset(&piTCP,0,sizeof(PROCESS_INFORMATION));
	sa.nLength = sizeof(SECURITY_ATTRIBUTES);
	sa.lpSecurityDescriptor = NULL;
	sa.bInheritHandle = TRUE;

	/* now get all the events, mutexes, etc */

	if (fDebug)
		WriteDebugLog("In RunLocalMPIJob");
	if (1 < JobInfo->Machine[JobInfo->dwMyMachineID].dwNumProcesses)
	{
		if (fDebug)
			WriteDebugLog("Creating a barrier semaphore");
		JobInfo->hBarrierSemaphore = CreateSemaphore(&sa,				/* security */
													 JobInfo->Machine[JobInfo->dwMyMachineID].dwNumProcesses - 1,	/* initial count */
													 JobInfo->Machine[JobInfo->dwMyMachineID].dwNumProcesses - 1,	/* max count */
													 NULL);				/* no name */
		_ASSERTE(JobInfo->hBarrierSemaphore);
		if (NULL == JobInfo->hBarrierSemaphore)
		{
			if (fDebug)
				WriteDebugLog("Unable to create Barrier Semaphore!  Exitting...\n");
			goto reporterror;
		}
		if (fDebug)
			WriteDebugLog("Creating a barrier event");
		JobInfo->hBarrierEvent = CreateEvent(&sa,	/* security */
											 TRUE,	/* manual reset */
											 FALSE, /* initially signalled? */
											 NULL); /* no name */
		_ASSERTE(JobInfo->hBarrierEvent);
		if (NULL == JobInfo->hBarrierEvent)
		{
			if (fDebug)
				WriteDebugLog("Unable to create Barrier Event!  Exitting...\n");
			goto reporterror;
		}
	}

	if (fDebug)
		WriteDebugLog("Creating a global terminator event");
	JobInfo->hGlobalTerminatorEvent = CreateEvent(&sa,		/* security */
												  TRUE,		/* manual reset */
												  FALSE,	/* initally signalled? */
												  NULL);	/* no name */
	_ASSERTE(JobInfo->hGlobalTerminatorEvent);
	if (NULL == JobInfo->hGlobalTerminatorEvent)
	{
		printf("Unable to create Global Terminator Event!  Exitting...\n");
		goto reporterror;
	}

	dwBufSize = sizeof(MPID_WINNT_SHMEM_BUFFER_T);
	for (i=0; i < JobInfo->Machine[JobInfo->dwMyMachineID].dwNumProcesses; i++)
	{
		if (fDebug)
		{
			sprintf(szMsg,"For local process [%d]",i);
			WriteDebugLog(szMsg);
			WriteDebugLog("\tCreating shared memory buffer");
		}
		JobInfo->Machine[JobInfo->dwMyMachineID].Process[i].hSharedMemoryBuffer = CreateFileMapping((HANDLE) 0xFFFFFFFF,		/* create a new one */
																									&sa,						/* null security attributes */
																									PAGE_READWRITE | SEC_NOCACHE | SEC_COMMIT,	/* read and write to the memory */
																									0,						/* high order bits */
																									dwBufSize,				/* low order bits */
																									NULL);					/* no name */
		_ASSERTE(JobInfo->Machine[JobInfo->dwMyMachineID].Process[i].hSharedMemoryBuffer);
		if (NULL == JobInfo->Machine[JobInfo->dwMyMachineID].Process[i].hSharedMemoryBuffer)
		{
			if (fDebug)
				WriteDebugLog("Unable to create Shared Memory Buffer for Message Passing!  Exitting...\n");
			goto reporterror;
		}

		if (fDebug)
			WriteDebugLog("\tMapping view of file");
		pBuf = (MPID_WINNT_SHMEM_BUFFER_T *)MapViewOfFile(JobInfo->Machine[JobInfo->dwMyMachineID].Process[i].hSharedMemoryBuffer, /* handle to file mapping */
														  FILE_MAP_ALL_ACCESS,											   /* all access */
														  0,															   /* high order bits */
														  0,															   /* low order bits */
														  0);															   /* map the whole file */
		_ASSERTE(pBuf);
		if (NULL == pBuf)
		{
			if (fDebug)
				WriteDebugLog("Unable to Map View of Shared Memory Buffer!  Exitting...\n");
			goto reporterror;
		}
		if (fDebug)
			WriteDebugLog("\tCreating buffer empty event");
		pBuf->hBufferEmpty = CreateEvent(&sa,
										 FALSE,
										 TRUE,
										 NULL);
		_ASSERTE(pBuf->hBufferEmpty);
		if (NULL == pBuf->hBufferEmpty)
			goto reporterror;
		if (fDebug)
			WriteDebugLog("\tCreating buffer full event");
		pBuf->hBufferFull = CreateEvent(&sa,
										FALSE,
										FALSE,
										NULL);
		_ASSERTE(pBuf->hBufferFull);
		if (NULL == pBuf->hBufferFull)
			goto reporterror;

		if (fDebug)
			WriteDebugLog("\tCreating mutex");
		pBuf->hMutex = CreateMutex(&sa,
								   FALSE,
								   NULL);
		_ASSERTE(pBuf->hMutex);
		if (NULL == pBuf->hMutex)
			goto reporterror;

		UnmapViewOfFile(pBuf);
	}
	
	
	ZeroMemory(&si, sizeof(si));
	si.cb = sizeof(si);

	printf("\n");

	for (i = 0; i < MAX_NUM_SHMEM_PROCS; i++)
	{
		LocalProcessHandles->hProc[i] = (HANDLE)NULL;
		LocalProcessHandles->hThread[i] = (HANDLE)NULL;
	}

	if (fDebug)
		WriteDebugLog("About to start processes");
	LocalProcessHandles->dwNumHandles = JobInfo->Machine[JobInfo->dwMyMachineID].dwNumProcesses;
	JobInfo->hLocalProcessHandles = hLocalProcessHandles;
	for (i=0; i < WorldProcessMap->dwNumProcs; i++)
	{
		if (WorldProcessMap->Process[i].dwMachineNumber == JobInfo->dwMyMachineID)
			WorldProcessMap->Process[i].eProtocol = E_SHMEM_PROTOCOL;
		else
			WorldProcessMap->Process[i].eProtocol = E_TCPIP_PROTOCOL;
	}
	
	/* FGG */
	if (m_hFile != NULL)
	{
		si.dwFlags=STARTF_USESTDHANDLES;
		si.hStdInput=NULL;
//		si.hStdOutput=m_hFile;
		si.hStdError=m_hFile;

		printf("Error output will be saved in file specified in 'mpirun.fgg'!\n\n");
	}
	/* End FGG */

	for (i=0; i < JobInfo->Machine[JobInfo->dwMyMachineID].dwNumProcesses; i++)
	{
		sprintf(szArgs," %s %d %s %d %s %d %s %d",
					   CMD_GLOBALRANK_STR,
					   JobInfo->Machine[JobInfo->dwMyMachineID].Process[i].dwGlobalRank,
					   CMD_JOBINFO_STR,
					   hJobInfoFileMap,
					   CMD_WORLDMAP_STR,
					   hWorldProcessMapFileMap,
					   CMD_PROCESS_HANDLES_STR,
					   hLocalProcessHandles);
		if (1 < JobInfo->dwNumMachines)
		{
			strcat(szArgs," ");
			strcat(szArgs,CMD_TCPIPPROC_STR);
		}
		strcat(JobInfo->Machine[JobInfo->dwMyMachineID].Process[i].szExecutable,szArgs);
		
		fStatus = CreateProcess(NULL,
								JobInfo->Machine[JobInfo->dwMyMachineID].Process[i].szExecutable,
								&sa,
								&sa,
								TRUE,
								CREATE_SUSPENDED,
								NULL,
								NULL,
								&si,
								&(piProcess[i]));
			
		_ASSERTE(fStatus);
		if (FALSE == fStatus)
			goto reporterror;
		_ASSERTE(piProcess[i].hProcess);
		LocalProcessHandles->hProc[i] = piProcess[i].hProcess;
		_ASSERTE(piProcess[i].hThread);
		LocalProcessHandles->hThread[i] = piProcess[i].hThread;
		if (fDebug)
		{
			sprintf(szMsg,"Created Process [%s] with ID [0x%x]...",JobInfo->Machine[JobInfo->dwMyMachineID].Process[i].szExecutable,piProcess[i].dwProcessId);
			printf("%s\n",szMsg);
			WriteDebugLog(szMsg);
		}
	}
	if (1 < JobInfo->dwNumMachines)
	{
		char szTCP[256];

		sprintf(szTCP,"mpitcpproc.exe %s %d %s %d %s %d %s %d %s %s",
				CMD_GLOBALRANK_STR,
				-1,
				CMD_JOBINFO_STR,
				hJobInfoFileMap,
				CMD_WORLDMAP_STR,
				hWorldProcessMapFileMap,
				CMD_PROCESS_HANDLES_STR,
				hLocalProcessHandles,
				CMD_TCPIPPROC_STR,
				CMD_TCPIPPROC_NOTIFY_STR);

		fStatus = CreateProcess(NULL,
								szTCP,
								&sa,
								&sa,
								TRUE,
								0,
								NULL,
								NULL,
								&si,
								&piTCP);
		
		_ASSERTE(fStatus);
		if (FALSE == fStatus)
			goto reporterror;
		if (fDebug)
		{
			sprintf(szMsg,"Created TCPIP agent [%s]\n",szTCP);
			WriteDebugLog(szMsg);
		}
	}

cleanup:
	return dwStatus;

reporterror:
	dwStatus = GetLastError();
	sprintf(szMsg,"\nExitting with status [%d]",dwStatus);
	printf("%s\n",szMsg);
	if (fDebug)
		WriteDebugLog(szMsg);

	goto cleanup;
}