#include "windows.h"
#include "packets.h"
#include "processinfo.h"
#include "infoprint.h"
#include "mpirun.h"
#include "crtdbg.h"

BOOL fExecutableSpecified = FALSE;
BOOL fDebug = FALSE;
DWORD dwNumProcs = 0;
DWORD dwMyMachineID = 0;
WORD wPortNumber = 0;
char szExecutable[MAXPATHLEN];
char szConfigFile[MAXPATHLEN];
char szMachineName[MAXPATHLEN];
BOOL fWait = TRUE;


DWORD main(int argc,char *argv[])
{
	DWORD dwStatus = ERROR_SUCCESS;
	int i;
	WORD wVersionRequested = MAKEWORD(1,1);
	WSADATA wsaData;
	BOOL wsaInit=FALSE;
	int nPlaceHolder = 3;

	dwStatus = (DWORD)WSAStartup(wVersionRequested,&wsaData);
	if (SOCKET_ERROR == (int)dwStatus)
		goto cleanup;
	wsaInit = TRUE;

	/*
		Leave WSA started until the end of this function.
		All functions called between here and the cleanup
		can assume that WSA has already been started
	*/

	if (2 > argc)
		goto printusage;

	for (i=0; i < argc; i++)
	{
		if (!strcmp(CMD_NOWAIT_STR,argv[i]))
		{
			fWait = FALSE;
			break;
		}
	}

	if ((nPlaceHolder + 1) <= argc)
	{
		if (!strcmp(CMD_PORTNUMBER_STR,argv[nPlaceHolder]))
		{
			wPortNumber = (WORD)atol(argv[nPlaceHolder+1]);
			nPlaceHolder += 2;
		}
	}

	if (!strcmp(CMD_NUMPROCS_STR,argv[1]))
	{
		if (4 > argc)
			goto printusage;
		dwNumProcs = (DWORD)atol(argv[2]);
		if (MAX_NUM_SHMEM_PROCS < dwNumProcs)
			goto printusage;
		strcpy(szExecutable,argv[nPlaceHolder]);
		for (i = nPlaceHolder+1; i < argc; i++)
		{
			strcat(szExecutable," ");
			strcat(szExecutable,argv[i]);
		}
		fExecutableSpecified = TRUE;
		dwStatus = RunMPIJob(argc,argv);
	}
	else if (!strcmp(CMD_CONFIGFILE_STR,argv[1]))
	{
		if (3 > argc)
			goto printusage;
		strcpy(szConfigFile,argv[2]);
		fExecutableSpecified = FALSE;
		dwStatus = RunMPIJob(argc,argv);
	}
	else if (!strcmp(CMD_SINGLE_PROC_STR,argv[1]))
	{
		if (4 > argc)
			goto printusage;
		strcpy(szMachineName,argv[2]);
		strcpy(szExecutable,argv[nPlaceHolder]);
		dwStatus = RunSingleProc(argc,argv);
	}
	else if (!strcmp(CMD_KILL_SINGLE_PROC_STR,argv[1]))
	{
		if (4 > argc)
			goto printusage;
		strcpy(szMachineName,argv[2]);
		dwStatus = KillSingleProc(argc,argv);
	}
	else
	{
		dwStatus = (DWORD)0xFFFFFFFF;
		goto printusage;
	}

	if (!strcmp(CMD_DEBUG_STR,argv[argc-1]))
		fDebug = TRUE;

cleanup:
	if (wsaInit)
		WSACleanup();
	printf("MPIRUN exited with status [%d]\n",dwStatus);
	return dwStatus;

printusage:
	dwStatus = (DWORD)0xFFFFFFFF;
	printf("\n");
	printf("MPIRUN is the start-up mechanism for shared-memory MPI programs\n");
	printf("Usage:\n");
    printf("\tmpirun -np <number of processes> [-p <portnumber>] <executable name> <executable args>\n");
	printf("\tmpirun -f  <configuration file> [-p <portnumber>]\n");
	printf("\tmpirun -s  <machine name> [-p <portnumber>] <executable name>\n");
	printf("\tmpirun -ks <machine name> [-p <portnumber>] <process handle>\n");
	printf("\tmpirun -kj <configuration file> [-p <portnumber>] <JOB ID>\n");
	printf("\n");
	printf("Currently, the maximum number of shared-memory process is limited to [%d]\n",
		   MAX_NUM_SHMEM_PROCS);
	printf("\n");
	printf("If both file 'mpirun.fgg' and 'mpirun.exe' are in the same directory,\n");
	printf("then the error output will be saved in file, specified in 'mpirun.fgg'.\n");
	printf("This addition was made by Grigori G.Fursin for his diploma project.\n");
	printf("\n");
	goto cleanup;
}
