#include <windows.h>
#include <windowsx.h>
#include <stdio.h>
#include <stdlib.h>
#include <process.h>
#include <tchar.h>
#include <string.h>
#include <ctype.h>
#include <winsock.h>
#include <malloc.h>
#include <crtdbg.h>
#include "service.h"
#include "debuglog.h"
#include "packets.h"



DWORD GetJobInfo(SOCKET hSocket,JOBINFO JobInfo[MAXPROCS],DWORD dwNumMachines)
{
	DWORD i,j;
	DWORD dwStatus = ERROR_SUCCESS;
	MPIServePacket_T CommandPacket;
	extern BOOL fDebugCmds;
	
	if (fDebugCmds)
	{
		char szMsg[256];
		WriteDebugLog("Entering GetJobInfo\n");
		sprintf(szMsg,"There are [%d] machines in this job\n",dwNumMachines);
		WriteDebugLog(szMsg);
	}
	for (i=0;(i < dwNumMachines) && (ERROR_SUCCESS == dwStatus);i++)
	{
		dwStatus = RecvSocketN(hSocket,(BYTE*)&CommandPacket,sizeof(MPIServePacket_T));
		if (SOCKET_ERROR == dwStatus)
		{
			dwStatus = WSAGetLastError();
			goto cleanup;
		}
		if (sizeof(MPIServePacket_T) == dwStatus)
			dwStatus = ERROR_SUCCESS;
		if (ERROR_SUCCESS == dwStatus)
		{
			MACHDATA *pMD;

			pMD = &(CommandPacket.uInfo.MachData);
			strcpy(JobInfo[i].szMachineName,pMD->szMachineName);
			JobInfo[i].dwNumExecutables = pMD->dwNumExecutables;
			JobInfo[i].GlobalRank = (DWORD *)malloc(pMD->dwNumExecutables * sizeof(DWORD));
			if (!(JobInfo[i].GlobalRank))
			{
				dwStatus = GetLastError();
				goto cleanup;
			}
			JobInfo[i].Executable = (char **)malloc(pMD->dwNumExecutables * sizeof(char*));
			if (!(JobInfo[i].Executable))
			{
				dwStatus = GetLastError();
				goto cleanup;
			}
			if (fDebugCmds)
			{
				char szMsg[256];
				sprintf(szMsg,"Num executables [%d]\n",JobInfo[i].dwNumExecutables);
				WriteDebugLog(szMsg);
				sprintf(szMsg,"Machine name [%s]\n",JobInfo[i].szMachineName);
				WriteDebugLog(szMsg);
			}
			for (j=0;j < JobInfo[i].dwNumExecutables; j++)
			{
				dwStatus = RecvSocketN(hSocket,(BYTE*)&CommandPacket,sizeof(MPIServePacket_T));
				if (SOCKET_ERROR == dwStatus)
				{
					dwStatus = WSAGetLastError();
					goto cleanup;
				}
				if (sizeof(MPIServePacket_T) == dwStatus)
					dwStatus = ERROR_SUCCESS;
				{
					PROCDATA *pPD;
					pPD = &(CommandPacket.uInfo.ProcData);
					if (fDebugCmds)
					{
						char szMsg[256];
						sprintf(szMsg,"Global rank [%d]\n",pPD->GlobalRank);
						WriteDebugLog(szMsg);
						sprintf(szMsg,"Executable [%s]\n",pPD->szExecutable);
						WriteDebugLog(szMsg);
					}
					JobInfo[i].GlobalRank[j] = pPD->GlobalRank;
					JobInfo[i].Executable[j] = strdup(pPD->szExecutable);
					if (!(JobInfo[i].Executable[j]))
					{
						dwStatus = GetLastError();
						goto cleanup;
					}
				}
			}
		}
	}

cleanup:
	if (fDebugCmds)
	{
		char szMsg[256];
		sprintf(szMsg,"Leaving GetJobInfo -- status [%d]\n",dwStatus);
		WriteDebugLog(szMsg);
	}
	return dwStatus;
}

VOID DBGPrintEnvString(char *pSSZ)
{
	extern BOOL fDebugCmds;
	if (fDebugCmds && pSSZ)
	{
		char *pTmp = pSSZ;
		char szMsg[256];
		while (pTmp)
		{
			if (strlen(pTmp))
			{
				sprintf(szMsg,"%s\n",pTmp);
				WriteDebugLog(szMsg);
				pTmp += (strlen(pTmp) + 1);
			}
			else
				pTmp = NULL;
		}
	}
}

void GetServicePath(char szPath[])
{
	HKEY hkLocalMachineSubKey;
	int nStatus;
	LPTSTR lpszLocalMachineSubKey = TEXT("System\\CurrentControlSet\\Services\\MPIServe");
	char szImagePath[256];

	nStatus = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
							lpszLocalMachineSubKey,
							0,
							KEY_ALL_ACCESS,
							&hkLocalMachineSubKey);
	if (ERROR_SUCCESS == nStatus)
	{
		LPTSTR lpszImagePath = TEXT("ImagePath");
		DWORD dwType;
		DWORD dwSize=256;

		nStatus = RegQueryValueEx(
					hkLocalMachineSubKey,
					lpszImagePath,
					0,
					&dwType,
					(BYTE*)szImagePath,
					&dwSize);
	}
	if (ERROR_SUCCESS == nStatus)
	{
		char *szTmp;

		szTmp = strrchr(szImagePath,'\\');
		if (szTmp)
			szTmp[0] = 0;
		strcpy(szPath,szImagePath);
	}
}


char *CopyEnvironment(char *pCopy,char *pOrig)
{
	char *pTmpCopy = 0;
	char *pTmpOrig = 0;

	if (pCopy && pOrig)
	{
		pTmpCopy = pCopy;
		pTmpOrig = pOrig;

		while (pTmpOrig[0] || pTmpOrig[1])
		{
			*pTmpCopy++ = *pTmpOrig++;
		}
		pTmpCopy[0] = pTmpCopy[1] = 0;
		pTmpCopy++;
	}

	return pTmpCopy;
}

DWORD StartJob(MPIServeThreadPacket_T *pPacket,JOBINFO JobInfo[MAXPROCS],DWORD dwNumMachines,DWORD dwMyID,SOCKET hSocket)
{
	LONG i, j;
	LONG bufsize;
	LONG errcode;
	LONG found;
	LONG numProcs;
	LONG nTCPIPProc = 0;
	LONG MPID_WorldSize = 0;
	char mapName[128];
	char tmpCmdLine[256];
	char szCmdLine[2048];
	char szTmpCmdLine[2048];
	char szMsg[256];
	char MyEnvStr[16364];
	char *pEnvStr;
	char *pssz;
	DWORD dwStatus = ERROR_SUCCESS;
	DWORD dwPortNumber = 12396;
	DWORD dwTotalProcs = 0;
	BOOL bStatus;
	BOOL created;
	BOOL TCPIPProcNeeded = FALSE;
	extern BOOL fDebugCmds;
	STARTUPINFO si;
	SECURITY_ATTRIBUTES saProcess;
	SECURITY_ATTRIBUTES saThread;
	SECURITY_ATTRIBUTES saFileMap;
	SECURITY_ATTRIBUTES saSem;
	SECURITY_ATTRIBUTES saMutex;
	SECURITY_ATTRIBUTES saEvent;
	
	HANDLE* hFileMaps;
	HANDLE* hProcs;
	HANDLE hBarrierEvent;
	HANDLE hBarrierSemaphore;
	HANDLE hTerminatorEvent;
	PROCESS_INFORMATION piProcess;

	MPID_WINNT_SHMEM_BUF_T** ShmemBufsView;

	found = 0;
	numProcs = JobInfo[dwMyID].dwNumExecutables;
	szCmdLine[0] = szTmpCmdLine[0] = '\0';
	pEnvStr = GetEnvironmentStrings();
	if (0 == pEnvStr)
		goto cleanup;
	pssz = CopyEnvironment(MyEnvStr,pEnvStr);
	FreeEnvironmentStrings(pEnvStr);
	sprintf(pssz,"PORTNUMBER=%d",pPacket->CommandPacket.uInfo.InitData.dwPortNumber);
	pssz += (strlen(pssz) + 1);
	pssz[0] = 0;		
	sprintf(pssz,"NUMPROCS=%d",JobInfo[dwMyID].dwNumExecutables);
	pssz += (strlen(pssz) + 1);
	pssz[0] = 0;
	sprintf(pssz,"MACHINEID=%d",dwMyID);
	pssz += (strlen(pssz) + 1);
	pssz[0] = 0;
	sprintf(pssz,"NUMMACHINES=%d",dwNumMachines);
	pssz += (strlen(pssz) + 1);
	pssz[0] = 0;
	for (i=0;i < (LONG)dwNumMachines;i++)
	{
		int j;
		sprintf(pssz,"MACHINE%d=%d",i,JobInfo[i].dwNumExecutables);
		pssz += (strlen(pssz) + 1);
		pssz[0] = 0;
		sprintf(pssz,"MACHINE%dNAME=%s",i,JobInfo[i].szMachineName);
		pssz += (strlen(pssz) + 1);
		pssz[0] = 0;
		for (j=0;j<(int)JobInfo[i].dwNumExecutables;j++)
		{
			sprintf(pssz,"MACHINE%dEXECUTABLE%d=%s",i,j,JobInfo[i].Executable[j]);
			pssz += strlen(pssz) + 1;
			pssz[0] = 0;
		}
	}
	dwTotalProcs = 0;
	for (i=0;i<(LONG)dwNumMachines;i++)
		dwTotalProcs += JobInfo[i].dwNumExecutables;
	sprintf(pssz,"MPID_WorldSize=%d",dwTotalProcs);
	pssz += strlen(pssz) + 1;
	pssz[0] = 0;

	if (dwNumMachines > 1)
	{
		TCPIPProcNeeded = TRUE;
		WriteDebugLog("The TCPIP Process is needed\n");
		numProcs += 1;
	}
	else
		WriteDebugLog("No TCPIP Process is needed\n");

	/* Allocate memory for file mappings' handles, processes' handles, and views */ 
	hFileMaps = (HANDLE*) malloc(numProcs*sizeof(HANDLE));
	
	_ASSERTE(_CrtIsValidPointer(hFileMaps,
								numProcs*sizeof(HANDLE),
								TRUE));

	hProcs = (HANDLE*) malloc(numProcs*sizeof(HANDLE));

	_ASSERTE(_CrtIsValidPointer(hProcs,
								numProcs*sizeof(HANDLE),
								TRUE));

	ShmemBufsView = (MPID_WINNT_SHMEM_BUF_T**) malloc(numProcs*sizeof(MPID_WINNT_SHMEM_BUF_T*));

	_ASSERTE(_CrtIsValidPointer(ShmemBufsView,
								numProcs*sizeof(HANDLE),
								TRUE));

	sprintf(szMsg,"Sizeof(MPID_WINNT_SHMEM_BUF_T) [%d]\n",sizeof(MPID_WINNT_SHMEM_BUF_T));
	WriteDebugLog(szMsg);

	/* Save number of processes in the environment:
		 all children inherit the environment and can get MPI_WordlSize, as well
		 as the rest of necessary info out of their environments */

	/* The size of the buffer for every process:
		 all processes see each other's buffers */
	bufsize = sizeof(MPID_WINNT_SHMEM_BUF_T)*numProcs;

	saFileMap.nLength = sizeof(SECURITY_ATTRIBUTES);
	saFileMap.lpSecurityDescriptor = NULL;
	saFileMap.bInheritHandle = TRUE;

	sprintf(szMsg,"Creating the file mappings for [%d] procs...\n",numProcs);
	WriteDebugLog(szMsg);
	for (i=0; i < numProcs; i++)
	{
		/* Create file mappings and save their handles in the environment*/
		hFileMaps[i] = CreateFileMapping ((HANDLE) 0xFFFFFFFF,
											&saFileMap, 
											PAGE_READWRITE,
											0,
											bufsize,
											mapName);
		_ASSERTE(hFileMaps[i]);
		sprintf(pssz, "Proc_%d_bufferHandle=%d", i,hFileMaps[i]);
		pssz += strlen(pssz) + 1;
		pssz[0] = 0;
		/* Map file mapping into the process address space */
		ShmemBufsView[i] = (MPID_WINNT_SHMEM_BUF_T*) MapViewOfFile (hFileMaps[i], 
																	FILE_MAP_ALL_ACCESS,
																	0,
																	0,
																	0);
		_ASSERTE(ShmemBufsView[i]);
		/* The memory for the shared buffers is allocated here in function MapViewOfFile */
	
		/* Initialize security attributes for the mutexes and events (they're gonna be 
			the same for all of them */
		saMutex.nLength = sizeof(SECURITY_ATTRIBUTES);
		saMutex.lpSecurityDescriptor = NULL;
		saMutex.bInheritHandle = TRUE;
		saEvent.nLength = sizeof(SECURITY_ATTRIBUTES);
		saEvent.lpSecurityDescriptor = NULL;
		saEvent.bInheritHandle = TRUE;

		WriteDebugLog("Creating the mutexes and events needed...\n");

		for (j=0; j < numProcs; j++)
		{
			/* Create mutexes and two events for receive buffers of the i-th process */
			/* System objects without names are good */
			/*  Assign handles to the appropriate fields in the shared memory buffers;
				so, we don't have to pass them explicitely to the child processes */
			ShmemBufsView[i][j].hMutex = CreateMutex(&saMutex, FALSE, NULL);
			sprintf(szMsg,"ShmemBufsView[%d][%d].hMutex -- [%d]\n",i,j,ShmemBufsView[i][j].hMutex);
			WriteDebugLog(szMsg);
			_ASSERTE(ShmemBufsView[i][j].hMutex);
			/* Filled-buffer event - auto-reset, non-signalled initially */
			ShmemBufsView[i][j].hBufferFilled = CreateEvent(&saEvent, FALSE, FALSE, NULL);
			sprintf(szMsg,"ShmemBufsView[%d][%d].hBufferFilled -- [%d]\n",i,j,ShmemBufsView[i][j].hBufferFilled);
			WriteDebugLog(szMsg);
			_ASSERTE(ShmemBufsView[i][j].hBufferFilled);
			/* Emptied-buffer event - auto-reset, signalled initially */
			ShmemBufsView[i][j].hBufferEmptied = CreateEvent(&saEvent, FALSE, TRUE, NULL);
			sprintf(szMsg,"ShmemBufsView[%d][%d].hBufferEmptied -- [%d]\n",i,j,ShmemBufsView[i][j].hBufferEmptied);
			WriteDebugLog(szMsg);
			_ASSERTE(ShmemBufsView[i][j].hBufferEmptied);
			/*  The rest of the ShmemBufsView[i][j] is the 
				Packet object of the type MPID_PKT_T */
					/* Initialize buffer fields */
			ShmemBufsView[i][j].Packet.head.lrank = -1;
			ShmemBufsView[i][j].Packet.head.tag = -1;
			ShmemBufsView[i][j].Packet.head.len = -1;
			/* Mutex and event handles are already there; we can access them */
		}
	}
	
	/* Create a Global Semaphore and Barrier Event for MPID_Barrier function */

	saEvent.nLength = sizeof(SECURITY_ATTRIBUTES);
	saEvent.lpSecurityDescriptor = NULL;
	saEvent.bInheritHandle = TRUE;
	saSem.nLength = sizeof(SECURITY_ATTRIBUTES);
	saSem.lpSecurityDescriptor = NULL;
	saSem.bInheritHandle = TRUE;
	/*
		If the number of executables we must run *on this machine*
		is greater than 1, we have to have barrier things.
	*/
	if (1 < JobInfo[dwMyID].dwNumExecutables)
	{
		WriteDebugLog("Creating Global Barrier Event...\n");
		hBarrierEvent = CreateEvent(&saEvent, 
									TRUE, 
									FALSE, 
									"__GLOBAL_BARRIER_EVENT__");
		_ASSERTE(hBarrierEvent);
		WriteDebugLog("Creating Global Barrier Semaphore...\n");
		hBarrierSemaphore = CreateSemaphore(&saSem, 
											JobInfo[dwMyID].dwNumExecutables - 1, 
											JobInfo[dwMyID].dwNumExecutables - 1, 
											"__GLOBAL_BARRIER_SEMAPHORE__");
		_ASSERTE(hBarrierSemaphore);
	}

	/* Create a Global Terminator event */
	WriteDebugLog("Creating Global Terminator Event...\n");
	hTerminatorEvent = CreateEvent(&saEvent, 
									TRUE, 
									FALSE, 
									"__GLOBAL_TERMINATOR_EVENT__");
	_ASSERTE(hTerminatorEvent);

	ZeroMemory(&si, sizeof(si));
	si.cb = sizeof(si);
	saProcess.nLength = sizeof(SECURITY_ATTRIBUTES);
	saProcess.lpSecurityDescriptor = NULL;
	saProcess.bInheritHandle = TRUE;
	saThread.nLength = sizeof(SECURITY_ATTRIBUTES);
	saThread.lpSecurityDescriptor = NULL;
	saThread.bInheritHandle = TRUE;
	saSem.nLength = sizeof(SECURITY_ATTRIBUTES);
	saSem.lpSecurityDescriptor = NULL;
	saSem.bInheritHandle = TRUE;

	WriteDebugLog("Attempting to start jobs...\n");	
	DBGPrintEnvString(MyEnvStr);
	/* Append the rank to each processor's command line and create the processes */
	for (i=0; i < (int)(JobInfo[dwMyID].dwNumExecutables); i++)
	{
		sprintf(tmpCmdLine,"%s %ld 0",JobInfo[dwMyID].Executable[i],JobInfo[dwMyID].GlobalRank[i]);
		printf("%s\n",tmpCmdLine);
		sprintf(szMsg,"Attempting to start [%s]...\n",tmpCmdLine);
		WriteDebugLog(szMsg);

		hProcs[i] = 0;	
		created = CreateProcess(NULL,
								tmpCmdLine,
								&saProcess,
								&saThread,
								TRUE,
								0,
								MyEnvStr,
								NULL,
								&si,
								&piProcess);

		if(!created) {
			errcode = GetLastError();
			sprintf(szMsg,"Fail to create process [%s] because [%d]...\n",
					tmpCmdLine,errcode);
			WriteDebugLog(szMsg);
			WriteDebugLog("Aborting the program.\n");
			goto cleanup;
		}
		else
		{
			sprintf(szMsg,"Successfully created process [%s]\n",tmpCmdLine);
			WriteDebugLog(szMsg);
		}
		hProcs[i] = piProcess.hProcess;
		sprintf(pssz, "Process%dHandle=%d", i,hProcs[i]);
		pssz += strlen(pssz) + 1;
		pssz[0] = 0;
	}

	if (TCPIPProcNeeded)
	{
		char szTmp[256];
		GetServicePath(szTmp);
		sprintf(tmpCmdLine,"%s\\mpitcpproc.exe %ld 1",szTmp,JobInfo[dwMyID].dwNumExecutables);
		printf("%s\n",tmpCmdLine);
		sprintf(szMsg,"Attempting to start [%s]...\n",tmpCmdLine);
		WriteDebugLog(szMsg);
		
		hProcs[JobInfo[dwMyID].dwNumExecutables] = 0;
		created = CreateProcess(NULL,
								tmpCmdLine,
								&saProcess,
								&saThread,
								TRUE,
								0,
								MyEnvStr,
								NULL,
								&si,
								&piProcess);
		if(!created)
		{
			errcode = GetLastError();
			sprintf(szMsg,"Fail to create TCPIP process because [%d]...\n",
					errcode);
			WriteDebugLog(szMsg);
			WriteDebugLog("Aborting the program.\n");
			goto cleanup;
		}
		else
		{
			sprintf(szMsg,"Successfully created process [%s]\n",tmpCmdLine);
			WriteDebugLog(szMsg);
		}
		/*	the last position in the array is specified by
			JobInfo[0].dwNumExecutables */
		hProcs[JobInfo[dwMyID].dwNumExecutables] = piProcess.hProcess;
	}

	/* Wait until the last one finishes and release memory */
	WriteDebugLog("Waiting for all processes to finish...\n");
	dwStatus = WaitForMultipleObjects(numProcs, 
										hProcs, 
										TRUE, 
										INFINITE);
	_ASSERTE(WAIT_FAILED != dwStatus);
	WriteDebugLog("All processes have finished\n");
	/* Dealocate all memory and handles */
	WriteDebugLog("Closing process handles...\n");
	for (i=0;i < (int)(JobInfo[dwMyID].dwNumExecutables); i++)
	{
		if (hProcs[i])
		{
			CloseHandle(hProcs[i]);
			hProcs[i] = 0;;
		}
	}
	if (TCPIPProcNeeded)
	{
		if (hProcs[JobInfo[dwMyID].dwNumExecutables])
		{
			CloseHandle(hProcs[JobInfo[dwMyID].dwNumExecutables]);
			hProcs[JobInfo[dwMyID].dwNumExecutables] = 0;
		}
	}

cleanup:
	WriteDebugLog("Cleaning up...\n");
	for (i=0;i < (int)(JobInfo[dwMyID].dwNumExecutables); i++)
	{
		if (hProcs[i])
		{
			sprintf(szMsg,"Having to kill process [%d]\n",i);
			WriteDebugLog(szMsg);
			if (TerminateProcess(hProcs[i],(UINT)-1))
				WriteDebugLog("Killed...\n");
		}
	}
	if (TCPIPProcNeeded)
	{
		if (hProcs[JobInfo[dwMyID].dwNumExecutables])
		{
			WriteDebugLog("Having to kill the TCP process\n");
			if (TerminateProcess(hProcs[JobInfo[dwMyID].dwNumExecutables],(UINT)-1))
				WriteDebugLog("Killed...\n");
		}
	}
	if (1 < JobInfo[dwMyID].dwNumExecutables)
	{
		WriteDebugLog("Closing Barrier Event and Semaphore...\n");
		bStatus = CloseHandle(hBarrierEvent);
		_ASSERTE(bStatus);
		bStatus = CloseHandle(hBarrierSemaphore);
		_ASSERTE(bStatus);
	}
	WriteDebugLog("Closing Terminator Event...\n");
	bStatus = CloseHandle(hTerminatorEvent);
	_ASSERTE(bStatus);

	WriteDebugLog("Closing shared memory buffers...\n");
	for (i=0; i < numProcs; i++) {
		for (j = 0; j < numProcs;j++) {
			bStatus = CloseHandle(ShmemBufsView[i][j].hMutex);
			_ASSERTE(bStatus);
			bStatus = CloseHandle(ShmemBufsView[i][j].hBufferFilled);
			_ASSERTE(bStatus);
			bStatus = CloseHandle(ShmemBufsView[i][j].hBufferEmptied);
			_ASSERTE(bStatus);
		}
		bStatus = UnmapViewOfFile(ShmemBufsView[i]);
		_ASSERTE(bStatus);
		bStatus = CloseHandle(hFileMaps[i]);
		_ASSERTE(bStatus);
		if (hProcs[i])
		{
			bStatus = CloseHandle(hProcs[i]);
			_ASSERTE(bStatus);
		}
	}
	if (TCPIPProcNeeded)
	{
		if (hProcs[JobInfo[dwMyID].dwNumExecutables])
		{
			bStatus = CloseHandle(hProcs[JobInfo[dwMyID].dwNumExecutables]);
			_ASSERTE(bStatus);
		}
	}
	WriteDebugLog("Done with job\n");
	return dwStatus;
}



DWORD WINAPI MPIServeStartJob(LPVOID pV)
{
	JOBINFO JobInfo[MAXPROCS];
	DWORD dwNumMachines = 0;
	DWORD dwMyID = 0;
	DWORD i,j;
	DWORD dwStatus = ERROR_SUCCESS;
	WORD wPortNumber = 0;
	MPIServeThreadPacket_T *pPacket = (MPIServeThreadPacket_T *)pV;
	SOCKET hSocket;
	MPIServePacket_T CommandPacket;
	extern BOOL fDebugCmds;

	if (fDebugCmds) {
		WriteDebugLog("In MPIServeStartJob\n");
	}
	if (0 == pPacket)
	{
		if (fDebugCmds)
			WriteDebugLog("Got NULL packet\n");
		goto cleanup;
	}
	CommandPacket = pPacket->CommandPacket;
	hSocket = pPacket->hSocket;
	memset(JobInfo,0,MAXPROCS * sizeof(JOBINFO));
	dwNumMachines = CommandPacket.uInfo.InitData.dwNumMachines;
	dwMyID = CommandPacket.uInfo.InitData.dwMachineID;
	wPortNumber = (WORD)CommandPacket.uInfo.InitData.dwPortNumber;
	if (fDebugCmds)
	{
		char szMsg[256];
		WriteDebugLog("Packet Information\n");
		sprintf(szMsg,"\teTask:\t\t\t[%d]\n",CommandPacket.eTask);
		WriteDebugLog(szMsg);
		sprintf(szMsg,"\tdwNumMachines:\t\t[%d]\n",CommandPacket.uInfo.InitData.dwNumMachines);
		WriteDebugLog(szMsg);
		sprintf(szMsg,"\tdwMyID:\t\t\t[%d]\n",CommandPacket.uInfo.InitData.dwMachineID);
		WriteDebugLog(szMsg);
		sprintf(szMsg,"\twPortNumber:\t\t[%d]\n",CommandPacket.uInfo.InitData.dwPortNumber);
		WriteDebugLog(szMsg);
		sprintf(szMsg,"\tszConfigFileName:\t[%s]\n",CommandPacket.uInfo.InitData.szConfigFileName);
		WriteDebugLog(szMsg);
	}
	if (ERROR_SUCCESS == dwStatus)
	{
		char szMsg[256];
		if (fDebugCmds) {
			WriteDebugLog("!!!!About to GetJobInfo!!!!\n");
		}
		dwStatus = GetJobInfo(hSocket,JobInfo,dwNumMachines);
		if (fDebugCmds)
		{
			sprintf(szMsg,"Status after GetJobInfo [%d]\n",dwStatus);
			WriteDebugLog(szMsg);
		}
	}

	if (fDebugCmds)
	{
		DWORD i,j;
		char szMsg[256];

		if (ERROR_SUCCESS == dwStatus)
		{
			dwStatus = OpenDebugLog("a+");
			for (i=0;(ERROR_SUCCESS == dwStatus) && (i<dwNumMachines);i++)
			{
				sprintf(szMsg,"The [%d] processes on machine [%s]:\n",JobInfo[i].dwNumExecutables,JobInfo[i].szMachineName);
				dwStatus = WriteDebugLog(szMsg);
				for (j=0;(ERROR_SUCCESS == dwStatus) && (j<JobInfo[i].dwNumExecutables);j++)
				{
					sprintf(szMsg,"\t[%s] with GlobalRank [%d]\n",JobInfo[i].Executable[j],JobInfo[i].GlobalRank[j]);
					dwStatus = WriteDebugLog(szMsg);
				}
				if (ERROR_SUCCESS == dwStatus)
					dwStatus = WriteDebugLog("\n");
			}
			CloseDebugLog();
			dwStatus = ERROR_SUCCESS;
		}
		else
		{
			WriteDebugLog("GetJobInfo failed!\n");
		}
	}

	/* Ok, we have all the info we need to start the job */
	if (ERROR_SUCCESS == dwStatus)
	{
		if (fDebugCmds) {
			WriteDebugLog("Attempting to start jobs\n");
		}
		dwStatus = StartJob(pPacket,JobInfo,dwNumMachines,dwMyID,hSocket);
	}
cleanup:
	for (i=0;i<dwNumMachines;i++)
	{
		if (JobInfo[i].Executable)
		{
			for (j=0;j < JobInfo[i].dwNumExecutables;j++)
			{
				if (JobInfo[i].Executable[j])
					free(JobInfo[i].Executable[j]);
			}
			free(JobInfo[i].Executable);
		}
		if (JobInfo[i].GlobalRank)
		{
			free(JobInfo[i].GlobalRank);
		}
	}
	if (pPacket)
	{
		if (pPacket->hSocket)
			closesocket(pPacket->hSocket);
		free(pPacket);
	}
	if (fDebugCmds)
	{
		char szMsg[256];

		sprintf(szMsg,"Status leaving MPIServeStartJob was [%d]\n",dwStatus);
		WriteDebugLog(szMsg);
	}
	return dwStatus;
}
