#include <stdio.h>
#include <windows.h>
#include <windowsx.h>
#include "packets.h"
#include "stdlib.h"
#include "time.h"
#include "queue.h"
#include "mpitcpproc.h"
#include "crtdbg.h"
#include "debuglog.h"


WINMPICHAPI JOBINFO *JobInfo;
WINMPICHAPI WORLDPROCESSMAP *WorldProcessMap;
WINMPICHAPI LOCALPROCESSHANDLES *LocalProcessHandles;

DWORD dwNumMachines = 0;
DWORD dwMyID = 0;
int rank = -1;
int world = -1;
char szConfigFileName[1024];
SOCKET sServerSocket = 0;


int PrintPacket(MPID_PKT_T *pPkt)
{
	printf("Packet:\n");
	printf("\tmode:\t\t[%d]\n",pPkt->head.mode);
	printf("\tsrc:\t\t[%d]\n",pPkt->head.src);
	printf("\tcontext_id:\t[%d]\n",pPkt->head.context_id);
	printf("\tlrank:\t\t[%d]\n",pPkt->head.lrank);
	printf("\ttag:\t\t[%d]\n",pPkt->head.tag);
	printf("\tlen:\t\t[%d]\n",pPkt->head.len);
	return 1;
}

BOOL GetCMDLineArgs(int argc,char *argv[])
{
	int i;
	BOOL fStatus = FALSE;

	for (i=0;(int)i < argc;i++)
	{
		if (!strcmp(argv[i],CONFIGFILEARG))
		{
			if (argv[i+1])
			{
				strcpy(szConfigFileName,argv[i+1]);
				break;
			}
		}
		else if (!strcmp(argv[i],SOCKETNUMARG))
		{
			sServerSocket = atoi(&(argv[i][strlen(SOCKETNUMARG)]));
			fStatus = TRUE;
		}
	}
	if (0 == sServerSocket)
	{
		WriteDebugLog("No socket specified!  Exitting...\n");
		fStatus = FALSE;
	}
	return fStatus;
}

int main(int argc, char** argv)
{
	MPID_PKT_T Pkt;
	MPID_PKT_T *pPkt = NULL;
	char *pBuffer = NULL;
	DWORD dwBufferSize = 0;
	DWORD dwStatus = ERROR_SUCCESS;
	WORD wPortNumber=0;
	BOOL fStatus=FALSE;
	HANDLE *hProcs=NULL;
	HANDLE hThread[2];
	extern HANDLE hReceiveThrough;
	DWORD i;
	JOBINFO *myJI = JobInfo;

	dwStatus = SetDebugFileName(DEBUGLOGFILENAME);

	pPkt = &Pkt;

	MPID_WINNT_init(&argc, &argv);

	dwNumMachines = JobInfo->dwNumMachines;
	wPortNumber = JobInfo->wPortNumber;
	dwMyID = JobInfo->dwMyMachineID;

	
	hProcs = (HANDLE*)malloc(JobInfo->Machine[dwMyID].dwNumProcesses * sizeof(HANDLE));
	if (NULL == hProcs)
	{
		dwStatus = GetLastError();
		goto cleanup;
	}
	for (i=0;i<JobInfo->Machine[dwMyID].dwNumProcesses;i++)
		hProcs[i] = LocalProcessHandles->hProc[i];

	/*
		Now that we have all the information we need,
		Open connections to all the other machines.
		To do this, wait for connections from all the machines
		numbered lower than you and then initiate connections
		for all machines greater than you.
	*/
	fStatus = EstablishTCPConnections(JobInfo,wPortNumber,&dwStatus);
	/*
		The shared memory listeners were set up in MPID_WINNT_Init.
		We should have all the sockets we need by now so we need to
		fire off the TCPIP sender/receiver threads and we should
		be on our way.
	*/
	if (fStatus)
	{
		/*
			The connections were made, fire off the listener and
			sender
		*/
		DWORD dwThreadID[2];
		extern DWORD WINAPI TCPIPSendThread(LPVOID);
		extern DWORD WINAPI TCPIPReceiveThread(LPVOID);

		hThread[0] = CreateThread(NULL,0,TCPIPSendThread,NULL,0,&(dwThreadID[0]));
		hThread[1] = CreateThread(NULL,0,TCPIPReceiveThread,NULL,0,&(dwThreadID[1]));
	}


	dwStatus = WaitForMultipleObjects(JobInfo->Machine[dwMyID].dwNumProcesses, LocalProcessHandles->hProc,TRUE,INFINITE);
	_ASSERTE(WAIT_FAILED != dwStatus);
	{
		BOOL fFinished;
		do
		{
			fFinished = IsEmptyQueue(SENDQUEUE) && IsEmptyQueue(RECEIVEQUEUE);
			Sleep(100);
		} while (!fFinished);
	}


	SetEvent(hEndProcessEvent);

	//WaitForMultipleObjects(2,hThread,TRUE,INFINITE);
	WaitForSingleObject(hThread[0],INFINITE);
	Sleep(1000);
	TerminateThread(hThread[1],0);
	/*
		Now the only thing to do is to wait until time to die,
		then wait until all the threads exit, then exit ourselves.
	*/

cleanup:
	MPID_WINNT_finalize();
	ExitProcess(dwStatus);
	return dwStatus;
}
